# MKS MFC Controller 647C DEV version 1.0 written by HR/JB@KIT 10/2013
# Rewritten for qkit/evaporate/deposition service HR,YS@KIT 2018
# Mass flow controller monitor/controller

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import qkit
from qkit.core.instrument_base import Instrument

import logging

import serial
import time,sys
import atexit

from threading import Lock


class MKS_647C(Instrument):
    """
    This is the driver for the MKS 647C four channel mass flow controller.

    Usage:
    Initialize with
    <name> = qkit.instruments.create('<name>', 'MKS_647C', address='<GBIP address>', reset=<bool>)
    """

    def __init__(self, name, address, reset=False):
        """
        Initializes the MKS, and communicates with the wrapper.
        
        Args:
            name (string)    : Name of the instrument
            address (string) : Serial address
            reset (bool)     : Resets to default values, default=False
        """
        
        logging.info(__name__ + ': Initializing instrument MKS_647C')
        Instrument.__init__(self, name, tags=['physical'])

        self.predef_pressure_unit = 13 # 100 ubar, depends on baratron
        self.predef_channels = {"Ar": 1, "N2": 2, "O2": 3, "ArO": 4}
        self.predef_gasCorfact = {"Ar": 1.37, "N2": 1.0, "O2": 1.0, "ArO": 1.37}  # Values as they are predefined in MKS
        self.predef_flowRange = {"Ar": 6, "N2": 6, "O2": 6, "ArO": 3}  # 100sccm, 100sccm, 100sccm, 10sccm

        self.mutex = Lock()
        self.setup(address)

        self.init_controller(reset=reset)
        
    def setup(self, address="COM10"):
        baudrate = 9600
        timeout = 1

        self.serial_port = self._std_open(address, baudrate, timeout)
        atexit.register(self.serial_port.close)
        
    def _std_open(self, device, baudrate, timeout):
        return serial.Serial(device, baudrate, timeout=timeout)

    def get_serial_port(self):
        return self.serial_port

    def get_mutex(self):
        return self.mutex

    def remote_cmd(self, cmd):
        # the 647C requires carriage return termination
        cmd += '\r'
        with self.mutex:
            self.serial_port.write(cmd)
        
            time.sleep(0.2)   #increase from .1 to .2
            #value = self.ser.readline().strip("\x06")
            rem_char = self.serial_port.inWaiting()
        
            value = self.serial_port.read(rem_char) # .strip("\x06")
            time.sleep(0.2)   #to avoid wrong communication
            return value.strip()

    """
    INSTRUMENT SETUP
    """
    def set_default(self):
        return self.remote_cmd("DF")

    def hardware_reset(self):
        return self.remote_cmd("RE")

    def get_version(self):
        return self.remote_cmd("ID")

    def init_controller(self, reset=False):
        """
        Initialize the MKS.
        Units and ranges required for value conversion are loaded and stored in self variables.

        If reset=True, the default values as defined by predef_... in __init__ are set and all channels are turned off.

        Args:
            reset: Reset the MKS to default values.

        Returns:
            Nothing but makes available:
            self.pressure_range, self.pressure_unit and self.flow_settings used internally for unit conversion.
        """
        print ("Initializing controller...")

        self.set_pressure_unit(self.predef_pressure_unit)
        self.pressure_range, self.pressure_unit = self.get_pressure_range()

        self.flow_settings = {}

        if reset:
            self.set_off_all()

        for chan_name in self.predef_channels.keys():
            channel = self.predef_channels[chan_name]
            self.flow_settings[channel] = {}
            if reset:
                self.set_off(channel)
                self.set_flow_range(channel, self.predef_flowRange[chan_name])
                self.set_gas_correction_factor(channel, self.predef_gasCorfact[chan_name])
            flow_range, flow_unit = self.get_flow_range(channel)
            gas_corr_fact = self.get_gas_correction_factor(channel)
            self.flow_settings[channel] = {"flow_range": float(flow_range),
                                           "flow_unit": str(flow_unit),
                                           "gas_corr_fact": float(gas_corr_fact)}
            if reset:
                self.set_flow(channel, 0)
        
        print ("Controller initialized.")
        if reset:
            print ("Values were resetted to predef values defined in __init__ and flow was turned off.")
        else:
            print ("No values were resetted.")
        print ("Channel settings are available as flow_settings[channel].")
        print ("Pressure settings: ")
        print (self.get_pressure_range())
        print ("Flow settings: ")
        for chan in self.flow_settings:
            print (chan, self.flow_settings[chan])

        return

    """
    CHANNEL SETUP
    """
    def set_mode(self, channel, mode=0, master_channel=0):
        """
        The Mode Selection defines the source of setpoint for each MFC channel. Possible modes are:
            - INDEP = independent
            - SLAVE = dependent to the actual flow of another channel
            - EXTERN = external source for setpoint
            - PCS = external controller
            - PID = built in PID controller
            - TEST = test for maintenance and installation
        (old name: setMode)

        Args:
            channel: Channel to be set.
            mode: Which mode to set
                0 = independent
                1 = slave
                2 = extern
                3 = PCS
                9 = test
            master_channel: Which channel to use as master in slave mode (1).
        """
        if mode == 1:
            cmd = "MO " + str(channel) + " 1 " + str(master_channel)
        else:
            cmd = "MO " + str(channel) + str(mode)
        return self.remote_cmd(cmd)

    def get_mode(self, channel):
        """
        Check for mode.
        (old name: getMode)

        Args:
            channel: Channel to be checked.

        Returns:
            mode, master channel
        """
        master_channel = -1
        cmd = "MO "+str(channel)+" R"
        mode, master_channel = (self.remote_cmd(cmd)).split()
        return int(mode), int(master_channel)

    def set_gas_correction_factor(self, channel, factor):
        """
        Set the gas correction factor for a given channel.

        Args:
            channel: Channel for which the gas correction factor is to be set.
            factor: The gas correction factor to be set.
        
        Note:
            This changes the gas settings to "USER, since the driver has no option to choose the gas directly.
            Chosing Argon directly from the menu on the controller gives a correction factor of 1.37.
            Nitrogen and Oxygen both are set to 1 by the device.
            This is programmed here in the predef vales of __init__.
        """
        cmd = "GC " + str(channel) + " " + str(int(factor*100))# '{fact:04d}'.format(fact=int(factor*100))
        ret = self.remote_cmd(cmd)
        gas_corr_fact = self.get_gas_correction_factor(channel)
        self.flow_settings[channel]["gas_corr_fact"] = float(gas_corr_fact)
        return ret

    def get_gas_correction_factor(self, channel):
        """
        Check gas correction factor of a channel.

        Args:
            channel: Channel to be checked.
        
        Returns:
            Gas correction factor.
        """
        cmd = "GC " + str(channel) + " R"
        res = self.remote_cmd(cmd)
        return float(res)/100.

    def set_flow_range(self, channel, value):
        """
        Set flow range for a channel.
        (old name: setFlowRange)

        Note:
            Most likely only 3 and 6 are used.

        Args:
            channel: Channel for which the flow range is to be set.
            value: 0..39
                0  = 1.000 SCCM
                1  = 2.000 SCCM
                2  = 5.000 SCCM
                3  = 10.00 SCCM
                4  = 20.00 SCCM
                5  = 50.00 SCCM
                6  = 100.0 SCCM
                7  = 200.0 SCCM
                8  = 500.0 SCCM
                9  = 1.000 SLM
                10 = 2.000 SLM
                11 = 5.000 SLM
                12 = 10.00 SLM
                13 = 20.00 SLM
                14 = 50.00 SLM
                15 = 100.0 SLM
                16 = 200.0 SLM
                17 = 400.0 SLM
                18 = 500.0 SLM
                19 = 1.000 SCMM
                20 = 1.000 SCFH
                21 = 2.000 SCFH
                22 = 5.000 SCFH
                23 = 10.00 SCFH
                24 = 20.00 SCFH
                25 = 50.00 SCFH
                26 = 100.0 SCFH
                27 = 200.0 SCFH
                28 = 500.0 SCFH
                29 = 1.000 SCFM
                30 = 2.000 SCFM
                31 = 5.000 SCFM
                32 = 10.00 SCFM
                33 = 20.00 SCFM
                34 = 50.00 SCFM
                35 = 100.0 SCFM
                36 = 200.0 SCFM
                37 = 500.0 SCFM
                38 = 30.00 SLM
                39 = 300.0 SLM
        """
        cmd = "RA " + str(channel) + " " + str(value)
        ret = self.remote_cmd(cmd)
        flow_range, flow_unit = self.get_flow_range(channel)
        self.flow_settings[channel]["flow_range"] = float(flow_range)
        self.flow_settings[channel]["flow_unit"] = str(flow_unit)
        return ret

    def set_flow_range_10sccm(self, channel):
        """
        Set flow range for a channel to 10 sccm.
        (old name: setFlowRange10sccm)

        Args:
            channel: Channel for which the flow range is to be set.
        """
        return self.set_flow_range(channel, 3)

    def set_flow_range_100sccm(self, channel):
        """
        Set flow range for a channel to 100 sccm.
        (old name: setFlowRange100sccm)

        Args:
            channel: Channel for which the flow range is to be set.
        """
        return self.set_flow_range(channel, 6)

    def get_flow_range(self, channel):
        """
        Check flow range of a channel.
        (old name: getFlowRange)

        Args:
            channel: Channel for which the flow range is to be checked.

        Reeturns:
            Flow range of the channel.
        """
        flows = [1.0, 2.0, 5.0, 10.0, 20.0, 50.0, 100.0, 200.0, 500.0,
                 1.0, 2.0, 5.0, 10.0, 20.0, 50.0, 100.0, 200.0, 400.0, 500.0,
                 1.0, 1.0, 2.0, 5.0, 10.0, 20.0, 50.0, 100.0, 200.0, 500.0,
                 1.0, 2.0, 5.0, 10.0, 20.0, 50.0, 100.0, 200.0, 500.0, 30.0, 300.0]
        units = ['SCCM', 'SCCM', 'SCCM', 'SCCM', 'SCCM', 'SCCM', 'SCCM', 'SCCM', 'SCCM',
                 'SLM', 'SLM', 'SLM', 'SLM', 'SLM', 'SLM', 'SLM', 'SLM', 'SLM', 'SLM',
                 'SCMM', 'SCFH', 'SCFH', 'SCFH', 'SCFH', 'SCFH', 'SCFH', 'SCFH', 'SCFH',
                 'SCFH', 'SCFM', 'SCFM', 'SCFM', 'SCFM', 'SCFM', 'SCFM', 'SCFM', 'SCFM',
                 'SCFM', 'SLM', 'SLM']

        cmd = "RA " + str(channel) + " R"
        flownum = int(self.remote_cmd(cmd))
        return flows[flownum], units[flownum]

    def set_high_limit(self, channel, limit):   #in 0.1 percent: 0..1100
        """
        Set an upper limit for the flow of a channel.
        (old name: setHighLimit)

        Args:
            channel: Channel for which the limit is to be set.
            limit: The limit to be set in 0.1 percent of full scale.
        """
        cmd = "HL " + str(channel) + " " + str(limit)
        return self.remote_cmd(cmd)

    def get_high_limit(self, channel):  # in 0.1 percent: 0..1100
        """
        Check for upper flow limit of a channel.
        (old name: getHighLimit)

        Args:
            channel: Channel for which the limit is to be checked.

        Returns:
            The set upper limit of the channel.
        """
        cmd = "HL " + str(channel) + " R"
        return self.remote_cmd(cmd)

    def set_low_limit(self, channel, limit):
        """
        Set a lower limit for the flow of a channel.
        (old name: setLowLimit)

        Args:
            channel: Channel for which the limit is to be set.
            limit: The limit to be set in 0.1 percent of full scale.
        """
        cmd = "LL " + str(channel) + " " + str(limit)
        return self.remote_cmd(cmd)
        
    def get_low_limit(self, channel):
        """
        Check for lower flow limit of a channel.
        (old name: getLowLimit)

        Args:
            channel: Channel for which the limit is to be checked.

        Returns:
            The set lower limit of the channel.
        """
        cmd = "LL " + str(channel) + " R"
        return self.remote_cmd(cmd)

    """
    PROCESS
    """
    def set_on_all(self):
        """
        Open main valve.
        (old name: setOnAll)

        Note:
            The channels in use have to be additionally turned on individually.
        """
        return self.remote_cmd("ON 0")

    def set_on(self, channel):
        """
        Open valve for a given channel.
        (old name: setOn)

        Note:
            The main valve has to be turned on separately by set_on_all.

        Args:
            channel: Channel to be turned on.
        """
        return self.remote_cmd("ON " + str(channel))

    def set_off_all(self):
        """
        Close main valve.
        (old name: setOffAll)

        Note:
            When the process is finished, the channels in use should be additionally turned off individually.
        """
        return self.remote_cmd("OF 0")

    def set_off(self, channel):
        """
        Close valve for a given channel.
        (old name: setOff)

        Note:
            The main valve can be turned off separately by set_off_all.

        Args:
            channel: Channel to be turned off.
        """
        return self.remote_cmd("OF " + str(channel))

    def get_channel_status(self, channel):
        """
        Check the status of a channel.
        (old name: check_channelStatus)

        Args:
            channel: Channel to be checked.

        Returns:
            Status bitstring with following bitwise information:
            0:  0/1 - channel off/on
            4:  trip limit low
            5:  trip limit high
            6:  overflow in
            7:  underflow in
            8:  overflow out
            9:  underflow out
            15: not used

        """
        return self.remote_cmd("ST " + str(channel))

    """
    FLOW
    """
    def get_flow(self, channel):
        """
        Check for actual flow of a channel.
        (old name: getActualFlow)

        Args:
            channel: Channel to be checked.

        Returns:
            Actual flow of a given channel in units defined by flow range.
        """
        cmd = "FL " + str(channel)
        flow = self.remote_cmd(cmd)
        return float(flow)*self.flow_settings[channel]['flow_range']*self.flow_settings[channel]['gas_corr_fact']/1000.

    def set_flow(self, channel, value):
        """
        Enter flow setpoint of a channel.
        (old name: setFlowSetPoint)

        Args:
            channel: Channel for which the flow is set.
            value: Flow to be set in units defined by flow range.
        """
        cmd = "FS " + str(channel) + str(int(round(value / self.flow_settings[channel]['flow_range'] /
                                                   self.flow_settings[channel]['gas_corr_fact']*1000.)))
        # print cmd
        return self.remote_cmd(cmd)

    def get_flow_setpoint(self, channel):
        """
        Check for flow setpoint of a channel.
        (old name: getFlowSetPoint)

        Args:
            channel: Channel for which the flow is set.
        
        Returns:
            Flow setpoint of a given channel in units defined by flow range.
        """
        cmd = "FS " + str(channel) +" R"
        ret = self.remote_cmd(cmd)
        # print ret
        return float(ret)*self.flow_settings[channel]['flow_range']*self.flow_settings[channel]['gas_corr_fact']/1000.

    """
    PRESSURE
    """
    def set_pressure_unit(self, pu=13):
        """
        Set the pressure unit.

        (old name: setPressureUnit)

        Args:
            pu: The number of the unit to be set.
                16: 10 mbar
                15: 1 mbar
                13: 100 ubar
                12: 10 ubar
                11: 1 ubar
        """
        cmd = "PU " + str(pu)
        ret = self.remote_cmd(cmd)
        self.pressure_range, self.pressure_unit = self.get_pressure_range()
        return ret

    def get_pressure_unit(self):
        """
        Get the pressure unit.
        (old name: getPressureUnit)

        Returns:
            The number of the unit set.
                16: 10 mbar
                15: 1 mbar
                13: 100 ubar
                12: 10 ubar
                11: 1 ubar
        """
        cmd = "PU R"
        return self.remote_cmd(cmd)
    
    def get_pressure_range(self):
        pressures = [1., 10., 100., 1000., 1., 10., 100., 1000., 1., 10., 100., 1., 10., 100., 1000., 1., 10., 100.,
                     1000., 1., 10., 100., 1., 10., 100., 1., 10., 100., 1000., 2., 5., 20., 50., 200., 500., 2000.,
                     5000., 2., 5., 20., 50., 200., 500., 2000., 5000., 2., 5., 20., 50., 2., 5., 20., 50., 200., 500.,
                     2000., 5000., 2., 5., 20., 50., 200., 500., 2000., 5000., 2., 5., 20., 50., 2., 5., 20., 50., 200.,
                     500., 2., 5., 20., 50., 200., 500., 2000., 5000., 1., 2., 5., 10]
        units = ['mTorr', 'mTorr', 'mTorr', 'mTorr', 'Torr', 'Torr', 'Torr', 'Torr', 'kTorr', 'kTorr', 'kTorr', 'uBar',
                 'uBar', 'uBar', 'uBar', 'mBar', 'mBar', 'mBar', 'mBar', 'Bar', 'Bar', 'Bar', 'Pa', 'Pa', 'Pa', 'kPa',
                 'kPa', 'kPa', 'kPa', 'mTorr', 'mTorr', 'mTorr', 'mTorr', 'mTorr', 'mTorr', 'mTorr', 'mTorr', 'Torr',
                 'Torr', 'Torr', 'Torr', 'Torr', 'Torr', 'Torr', 'Torr', 'kTorr', 'kTorr', 'kTorr', 'kTorr', 'ubar',
                 'ubar', 'ubar', 'ubar', 'ubar', 'ubar', 'ubar', 'ubar', 'mbar', 'mbar', 'mbar', 'mbar', 'mbar', 'mbar',
                 'mbar', 'mbar', 'bar', 'bar', 'bar', 'bar', 'Pa', 'Pa', 'Pa', 'Pa', 'Pa', 'Pa', 'kPa', 'kPa', 'kPa',
                 'kPa', 'kPa', 'kPa', 'kPa', 'kPa', 'Mpa', 'Mpa', 'Mpa', 'Mpa']
        pnum = int(self.get_pressure_unit())
        return pressures[pnum], units[pnum]

    def get_pressure(self):
        """
        Check for pressure.
        (old name: getActualPressure)

        Returns:
            Actual pressure.
            Converted from 0.1 percent of full scale.
        """
        cmd = "PR"
        return float(self.remote_cmd(cmd))*self.pressure_range/1000

    def set_pressure_mode(self, On=True):
        """
        Enter pressure mode (pressure controlled automatically).
        (old name: setPressureMode)

        Args:
            On (bool): True for on (auto), False for off.
        """
        if On:
            cmd = "PM 1"
        else:
            cmd = "PM 0"
        return self.remote_cmd(cmd)

    def get_pressure_mode(self):
        """
        Check for pressure mode (pressure controlled automatically).
        (old name: PressureModeOn)

        Returns:
            Pressure mode status (True for on (auto), False for off.
        """
        cmd = "PM R"
        status = self.remote_cmd(cmd)
        if int(status) == 1:
            return True
        else:
            return False

    def set_pressure(self, value):
        """
        Enter pressure setpoint.
        (old name: setPressureSetPoint)

        Args:
            Setpoint in 0.1 percent of full scale.
        """
        # value in 0.1 percent of fullscale: 0..1100
        cmd = "PS "+str(value)
        return self.remote_cmd(cmd)

    def get_pressure_setpoint(self):
        """
        Check for pressure setpoint.
        (old name: getPressureSetPoint)

        Returns:
            Pressure setpoint.
        """
        cmd = "PS R"
        return self.remote_cmd(cmd)

    def get_pcs(self):
        """
        Check for PCS in PCS mode for use with an external pressure controller (e.g. type 250).
        All  gas  flow  channels  which  are  configured  in  the  PCS  mode  are regulated through the
        pressure control signal (PCS) according to the ratio of their set points.
        (old name: getActualPCS)

        Returns:
            Actual PCS signal in 0.1 percent of full scale.
        """
        cmd = "PC"
        return self.remote_cmd(cmd)

if __name__ == "__main__":

    MKS = MKS_647C("MKS", address="COM10")

    MKS.init_controller()



# -*- coding: utf-8 -*-
"""
@author: S1@KIT/2020
Before we start, we check whether essential external libraries are available
and check for commonly used libraries.
"""

import qkit
from pkgutil import find_loader


class ModuleAvailable:
    """
    This class hosts and provides information about the installed python modules.
    To check wether a module is available, you can use getitem or the function call
    to the instance.
    Once the availability of a module is checked, the result is stored in a dictionary.
    print(instance) will give you all checked modules
    """
    
    def __init__(self):
        self.available_modules = {}
    
    def module_available(self, module_name):
        if module_name not in self.available_modules:
            self.available_modules[module_name] = bool(find_loader(module_name))
        return self.available_modules[module_name]
    
    def __call__(self, module_name):
        return self.module_available(module_name)
    
    def __getitem__(self, module_name):
        return self.module_available(module_name)
    
    def __repr__(self):
        return repr(self.available_modules)


qkit.module_available = ModuleAvailable()

# In qkit.cfg['blocked_modules'] you can specify modules you do not want to use.
# This can be helpful for programming and debugging but also if a module is not working properly.
for e in qkit.cfg.get('blocked_modules',[]):
    qkit.module_available.available_modules[e] = False


# These modules are essential.
# Qkit does not start if they are not available.
ESSENTIALS = ['numpy', 'h5py', 'zmq']

for e in ESSENTIALS:
    if not qkit.module_available(e):
        raise ImportError(
                "Module '%s' not found. The following modules are essentially needed for basic operation of QKIT:\n - %s" % (e, "\n - ".join(ESSENTIALS)))

# These modules are optional, but very common.
# Typically, they are not used for key functions, so the code should mostly work without them
OPTIONALS = ["IPython", "PyQt4", "PyQt5", "ipywidgets", "matplotlib", "pandas", "peakutils", "pyqtgraph", "scipy"]
for e in OPTIONALS:
    qkit.module_available(e)




from qkit.core.instrument_base import Instrument
import instruments
import types
import logging
import numpy as np
from time import sleep

class te_opto_dac(Instrument):
    '''
        This is the instrument driver for optically decoupled 
        Tunnelelektronik single- and triple channel DACs
    '''

    def __init__(self, name, source=None, channel = 0, dachannels = 1, inverted = False):
		Instrument.__init__(self, name, tags=['virtual'])
		self._instruments = instruments.get_instruments()
		self._source= self._instruments.get(source)
		self._channel = channel # channel = 'portn/linem'
		self._nchannels = dachannels
		self._values = np.zeros((dachannels))
		self._inverted = inverted
        
		# Add parameters        
		self.add_parameter('value', type=float,
			channels=(1, dachannels), channel_prefix = 'ch%d_',
			flags=Instrument.FLAG_SET, units='V')
		self.add_parameter('invert', type=bool,
			flags=Instrument.FLAG_SET, units='')
		#self.add_parameter('binary', type=int,
		#	flags=Instrument.FLAG_SET, units='')

    def get_all(self):
        pass
        
    def do_set_value(self, val, channel):
        ''' 
            output a float between -10. and 10. on the dac
        '''
        if((val > 10) || (val < -10)):
            raise ValueError('Valid outputs of TE DACs are be between -10V and +10V.')
        val = np.int16(min(2**15-1, max(-2**15, val/10*1**15)))
        self.do_set_binary(val, channel)
    def do_set_binary(self, val, channel):
        '''
            pass a 16bit signed integer sample to the dac
            
            format:
            000000000 --- (01x * 16 bits)) * nchannels --- 000000000
            where x are the inverted bits of the input
            
            original matlab code:
            disc_val=uint16(round(inp/10*65536));
            ofs=10
            a(2+ofs:3:16*3+ofs)=1-bitget(disc_val,16:-1:1); # split ~disc_val into bits
            a(1+ofs:3:16*3+ofs)=ones(1,16);
            a=[a zeros(1,9)];
        '''
        self._values[channel] = val

        # unpack bits
        bits = np.zeros((9), np.bool)
        for val in self._values:
            number_bits = [bool(np.int16(val) & 1<<n) for n in range(16, 0, -1)]
            code_bits = np.zeros(3*16, np.bool)
            code_bits[1:1+3*16:3] = np.ones((16), np.bool)
            code_bits[2:2+3*16:3] = number_bits
            bits = np.append(bits, code_bits)
        bits = np.append(bits, np.zeros((9), np.bool))
        if(self._inverted): bits = ~bits

        # push bits to device; matlab code outputs stream of bits at 19kS/s
        # variant 1: bit-bang
        # variant 2: hardware streaming
        self._source.digital_stream(self._channel, np.atleast_2d(bits).transpose(), rate = 19e3)
        #self._source.digital_out(self._channel, bla)



# nitrogen_controller.py by JB@KIT 04/2015 jochen.braumueller@kit.edu
# client script to be run on the control machine (Raspberry Pi) to monitor and control the server (nitrogen_server.py)
# use: execute after starting the server

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import socket
import sys
import time
import os.path

HOST, PORT = 'ip-address', 9989

def server_request(data_string):
    
    try:
        received = None
        # Create a socket (SOCK_STREAM means a TCP socket)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # Connect to server and send data
        sock.connect((HOST, PORT))
        sock.sendall(data_string + "\n")
        # Receive data from the server and shut down
        received = str(sock.recv(1024))
    finally:
        sock.close()
        return received
        

if __name__ == "__main__":

    if not os.path.isfile('monitor.log'):
        file = open('monitor.log','w')
        file.close()
        
    print 'Start nitrogen refilling controller...'
    with open('monitor.log','a') as f: f.write('\n' + time.ctime() + '   Start nitrogen refilling control script.\n')
    
    controller_status_hist = False
    manual_status = True
    t_start = time.time() + 24*3600
    while True:
        try:
            received = server_request('get_status')
            controller_status = bool(int(received[0]))
            if controller_status != controller_status_hist:   #status change
                s = time.ctime() + ':   status change   ' + received
                print s
                with open('monitor.log','a') as f: f.write(s + '\n')   #append to log
                controller_status_hist = controller_status
                
                if controller_status:   #switched on
                    t_start = time.time()   #reset timer

            if (time.time() - t_start) > 60 * 60 and controller_status and manual_status:   # >60min refilling
                x = server_request('output off')
                s = time.ctime() + ':   refilling longer than 60min -> switch off'
                print s
                with open('monitor.log','a') as f: f.write(s + '\n')   #append to log
                manual_status = False
            if not manual_status and not controller_status:   #when automation wants to switch off
                x = server_request('output on')
                s = time.ctime() + ':   automation switch off -> deactivate manual off'
                print s
                with open('monitor.log','a') as f: f.write(s + '\n')   #append to log
                manual_status = True
            time.sleep(10)
        except Exception as e:
            print e


# Agilent_E8257D.py class, to perform the communication between the Wrapper and the device
# Pieter de Groot <pieterdegroot@gmail.com>, 2008
# Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa

import logging
import numpy

class Keysight_E8267D(Instrument):
    '''
    This is the driver for the Keysight E8257D Signal Genarator

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Keysight_E8257D', address='<GBIP address>, reset=<bool>')
    '''

    def __init__(self, name, address, reset=False):
        '''
        Initializes the Keysight_E8267D, and communicates with the wrapper.

        Input:
          name (string)    : name of the instrument
          address (string) : GPIB address
          reset (bool)     : resets to default values, default=False
        '''
        logging.info(__name__ + ' : Initializing instrument Keysight_E8267D')
        Instrument.__init__(self, name, tags=['physical'])

        # Add some global constants
        self._address = address
        self._visainstrument = visa.instrument(self._address)

        self.add_parameter('power',
            flags=Instrument.FLAG_GETSET, units='dBm', minval=-135, maxval=25, type=float)
        self.add_parameter('phase',
            flags=Instrument.FLAG_GETSET, units='rad', minval=-numpy.pi, maxval=numpy.pi, type=float)
        self.add_parameter('frequency',
            flags=Instrument.FLAG_GETSET, units='Hz', minval=1e5, maxval=20e9, type=float)
        self.add_parameter('status',
            flags=Instrument.FLAG_GETSET, type=bool)

        self.add_function('reset')
        self.add_function ('get_all')


        if (reset):
            self.reset()
        else:
            self.get_all()
        
        if self._visainstrument.query("*IDN?").split(",")[1].strip() == 'N5173B':
            self.set_parameter_bounds("power",-20,19)
            self.set_parameter_bounds("frequency",9e3,20e9)
        else:
            print('this is a copy of the Agilent E8257D driver, thourough testing of this driver in combination with the Keysight E8267D has NOT been done! proceed with caution!')
            print('The max. output power and the frequency range depend on the order option of the Keysight E8267D! ')
    def reset(self):
        '''
        Resets the instrument to default values

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : resetting instrument')
        self._visainstrument.write('*RST')
        self.get_all()

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : get all')
        self.get_power()
        self.get_phase()
        self.get_frequency()
        self.get_status()

    def do_get_power(self):
        '''
        Reads the power of the signal from the instrument

        Input:
            None

        Output:
            ampl (?) : power in ?
        '''
        logging.debug(__name__ + ' : get power')
        return float(self._visainstrument.query('POW:AMPL?'))

    def do_set_power(self, amp):
        '''
        Set the power of the signal

        Input:
            amp (float) : power in ??

        Output:
            None
        '''
        logging.debug(__name__ + ' : set power to %f' % amp)
        self._visainstrument.write('POW:AMPL %s' % amp)

    def do_get_phase(self):
        '''
        Reads the phase of the signal from the instrument

        Input:
            None

        Output:
            phase (float) : Phase in radians
        '''
        logging.debug(__name__ + ' : get phase')
        return float(self._visainstrument.query('PHASE?'))

    def do_set_phase(self, phase):
        '''
        Set the phase of the signal

        Input:
            phase (float) : Phase in radians

        Output:
            None
        '''
        logging.debug(__name__ + ' : set phase to %f' % phase)
        self._visainstrument.write('PHASE %s' % phase)

    def do_get_frequency(self):
        '''
        Reads the frequency of the signal from the instrument

        Input:
            None

        Output:
            freq (float) : Frequency in Hz
        '''
        logging.debug(__name__ + ' : get frequency')
        return float(self._visainstrument.query('FREQ:CW?'))

    def do_set_frequency(self, freq):
        '''
        Set the frequency of the instrument

        Input:
            freq (float) : Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : set frequency to %f' % freq)
        self._visainstrument.write('FREQ:CW %s' % freq)

    def do_get_status(self):
        '''
        Reads the output status from the instrument

        Input:
            None

        Output:
            status (string) : 'On' or 'Off'
        '''
        logging.debug(__name__ + ' : get status')
        return bool(int(self._visainstrument.query('OUTP?')))


    def do_set_status(self, status):
        '''
        Set the output status of the instrument

        Input:
            status (string) : 'On' or 'Off'

        Output:
            None
        '''
        logging.debug(__name__ + ' : set status to %s' % status)
        
        if status == True:
            self._visainstrument.write('OUTP ON')
        elif status == False:
            self._visainstrument.write('OUTP OFF')
        else:
            raise ValueError('set_status(): can only set True or False')


    # shortcuts
    def off(self):
        '''
        Set status to 'off'

        Input:
            None

        Output:
            None
        '''
        self.set_status(False)

    def on(self):
        '''
        Set status to 'on'

        Input:
            None

        Output:
            None
        '''
        self.set_status(True)



from __future__ import annotations

from abc import ABC, abstractmethod
from functools import wraps
import inspect
from typing import Any, Callable, List
from enum import Enum, unique, auto
import numpy as np

from qkit.core.instrument_base import Instrument
from qkit.drivers.visa_prologix import instrument
import logging

class ModernInstrument(Instrument):
    """
    Base class for instruments.

    Use @QkitProperty() to convert a getter to a property.
    Use @$prop_name$.setter to add a setter to a property.
    Do note, that the function name of the getter and setter must be the same.  
    """
    FLAG_GET = 0x01             # Parameter is gettable
    FLAG_SET = 0x02             # {arameter is settable
    FLAG_GETSET = 0x03          # Shortcut for GET and SET
    FLAG_GET_AFTER_SET = 0x04   # perform a 'get' after a 'set'
    FLAG_SOFTGET = 0x08         # 'get' operation is simulated in software,
                                # e.g. an internally stored value is returned.
                                # Only use for parameters that cannot be read
                                # back from a device.

    class WrappedFunction:

        def __init__(self, func) -> None:
            self.func = func

        def call(self, *args, **kwargs):
            return self.func(*args, **kwargs)

        @classmethod
        def from_qkit_function(cls, target_object: Any, func: function) -> ModernInstrument.WrappedFunction:
            return cls(lambda *args, **kwargs: func(target_object, *args, **kwargs))

    class WrappedProperty:
        """
        A wrapper around QKit-Property with the object information required
        to call the functions on an object.
        Qkit-Property does not belong to an instance, but to an class.
        """

        def __init__(self, getter, setter, metadata: PropertyMetadata) -> None:
            self._getter = getter
            self._setter = setter
            self._metadata = metadata

        def build_options(self) -> dict:
            """
            Backwards compatibility to the old ways of QKit.
            """
            return {
                "set_func": self._setter,
                "get_func": self._getter,
                "flags": Instrument.FLAG_GETSET
            }

        def get(self, query=True, **kwargs):
            """
            Returns the property value.
            """
            return ModernInstrument._convert_value(self._getter(query=query, **kwargs), self._metadata.type)

        def set(self, value):
            """
            Sets the property value.
            """
            self._setter(value)

        def get_metadata(self) -> PropertyMetadata:
            """
            Returns the properties metadata
            """
            return self._metadata

        @classmethod
        def from_qkit_property(cls, target_object: Any, property: QkitProperty) -> ModernInstrument.WrappedProperty:
            """
            Converts a QKitProperty into a Wrapped Property to be handled in a ModernInstrument Instance.
            """
            return cls(lambda **kwargs: property._get_logic(target_object, **kwargs), lambda val: property.__set__(target_object, val), property.metadata)


    def __init__(self, name, **kwargs):
        self._name = name
        self._initialized = False
        self._instrument_options = kwargs
        self._tags = getattr(kwargs, "tags", [])
        self._parameters = {}
        self._parameter_groups = {}
        self._functions = {}
        self._options = {"tags": []}

    def discover_capabilities(self):
        """
        Call this method from init to self-discover QkitNative functions and properties.
        """
        visitor = QkitNativeVisitor(self)
        visitor.visit_instance(self)

    def __str__(self):
        return f"Instrument '{self._name}'"

    def get_name(self):
        "Return the name of the instrument"
        return self._name

    def get_type(self):
        """
        Return the type of the instrument as a string.
        Might behave unexpectedly in case of new drivers.
        """
        return str(self.__module__)

    def _json(self):
        '''
        This function is called by the QkitJSONEncoder and should return a 
        dictionary that will be saved to the JSON file.
        '''
        return {'dtype' : 'qkitInstrumentv2', 'content': str(self.get_name())}

    def get_options(self):
        """
        Return the user provided instrument options.
        """
        return self._instrument_options

    def get_tags(self):
        """
        Return the user provided instrument tags.
        """
        return self._tags

    def has_tag(self, tags):
        if type(tags) is list:
            for tag in tags:
                if tag in self._tags:
                    return True
            return False
        else:
            if tags in self._tags:
                return True

    def initialize(self):
        self._initialized = True

    def remove(self):
        self._remove_parameters()

    def is_initialized(self):
        return self._initialized

    def add_parameter(self, name, **kwargs):
        # FIXME: Reimplement via ObjectTreeVisitor
        raise NotImplementedError("add_parameter no longer supported in ModernInstrument!")
    
    def _remove_parameters(self):
        self._parameters = {}

    def remove_parameter(self, name):
        if name in self._parameters:
            del self._parameters[name]

    def has_parameter(self, name):
        return (name in self._parameters)

    def get_parameter_options(self, name):
        '''
        Return list of options for paramter.

        Input: name (string)
        Output: dictionary of options
        '''
        if name in self._parameters:
            return self._parameters[name].build_options()
        else:
            return None

    def get_shared_parameter_options(self, name):
        '''
        Return list of options for paramter.

        Input: name (string)
        Output: dictionary of options
        '''
        if name in self._parameters:
            options = dict(self._parameters[name].build_options())
            for i in ('get_func', 'set_func'):
                if i in options:
                    del options[i]
            if 'type' in options and options['type'] is type(None):
                options['type'] = None
            return options
        else:
            return None

    def set_parameter_options(self, name, **kwargs):
        '''
        Change parameter options.

        Input:  name of parameter (string)
        Ouput:  None
        '''
        if name not in self._parameters:
            logging.warn('Parameter %s not defined' % name)
            return None

        for key, val in kwargs.items():
            self._parameters[name][key] = val


    def get_parameter_tags(self, name):
        '''
        Return tags for parameter 'name'.

        Input:  name of parameter (string)
        Ouput:  array of tags
        '''

        if name not in self._parameters:
            return []

        return self._parameters[name].metadata.tags

    def add_parameter_tag(self, name, tag):
        '''
        Add tag to list of tags for parameter 'name'.

        Input:  (1) name of parameter (string)
                (2) tag (string)
        Ouput:  None
        '''

        if name not in self._parameters:
            return

        self._parameters[name]['tags'].append(tag)

    def set_parameter_bounds(self, name, minval, maxval):
        '''
        Change the bounds for a parameter.

        Input:  (1) name of parameter (string)
                (2) minimum value
                (3) maximum value
        Output: None
        '''
        self.set_parameter_options(name, minval=minval, maxval=maxval)

    def set_channel_bounds(self, name, channel, minval, maxval):
        '''
        Change the bounds for a channel.

        Input:  (1) name of parameter (string)
                (2) channel number (int)
                (3) minimum value
                (4) maximum value
        Output: None
        '''

        opts = self.get_parameter_options(name)
        if 'channel_prefix' in opts:
            var_name = opts['channel_prefix'] % channel + name
        else:
            var_name = '%s%d' % (name, channel)

        self.set_parameter_options(var_name, minval=minval, maxval=maxval)

    def set_parameter_rate(self, name, stepsize, stepdelay):
        '''
        Change the rate properties for a channel.

        Input:
            stepsize (float): the maximum step size
            stepdelay (float): the delay after each step
        Output:
            None
        '''
        self.set_parameter_options(name, maxstep=stepsize, stepdelay=stepdelay)

    def get_parameter_names(self):
        '''
        Returns a list of parameter names.

        Input: None
        Output: all the paramter names (list of strings)
        '''
        return self._parameters.keys()

    def get_parameters(self):
        '''
        Return the parameter dictionary.

        Input: None
        Ouput: Dictionary, keys are parameter names, values are the options.
        '''
        return self._parameters

    def get_shared_parameters(self):
        '''
        Return the parameter dictionary, with non-shareable items stripped.

        Input: None
        Ouput: Dictionary, keys are parameter names, values are the options.
        '''

        params = {}
        for key in self._parameters:
            params[key] = self.get_shared_parameter_options(key)
        return params

    def get_parameter_groups(self):
        '''
        Return a dictionary with parameter group name -> group members.
        '''
        return self._parameter_groups

    _CONVERT_MAP = {
            int: int,
            float: float,
            bytes: str,
            str: str,
            bool: bool,
            tuple: tuple,
            list: list,
            np.ndarray: lambda x: x.tolist(),
            np.int64: int,
            np.float64: float
    }

    @staticmethod
    def _convert_value(value, ttype):
        if type(value) is bool and \
                ttype is not bool:
            raise ValueError('Setting a boolean, but that is not the expected type')

        if ttype not in ModernInstrument._CONVERT_MAP:
            raise ValueError('Unsupported type %s', ttype)

        try:
            func = ModernInstrument._CONVERT_MAP[ttype]
            value = func(value)
        except:
            raise ValueError('Conversion of %r to type %s failed',value, ttype)

        return value

    def get(self, name, query=True, fast=False, **kwargs):
        try:
            return self._parameters[name].get(query=query, **kwargs)
        except (AttributeError, TypeError, NameError, ValueError) as e:
            logging.error(f"Caught error while processing {name}(query={query}): e")
            raise e
    
    def set(self, name, value, **kwargs):
        return self._parameters[name].set(value, **kwargs)

    def call(self, name, *args, **kwargs):
        return self._functions[name].call(*args, **kwargs)

def interval_check(lower, upper):
    """
    Return a callable which returns true if its argument lies between the provided lower and upper bound.
    """
    def checker(arg):
        return lower <= arg <= upper
    checker.lower = lower
    checker.upper = upper
    return checker


def QkitFunction(func):
    """
    A wrapper of functions to facilitate easy self-discovery.
    """
    func._qkit_function = True
    return func

@unique
class CachePolicy(Enum):
    SOFT_GET = auto()
    ALWAYS_REFRESH = auto()
    NONE = auto() # Default value

def caching(cache_policy: CachePolicy, get_after_set = False):
    def decorator(func):
        assert isinstance(func, QkitProperty)
        func.set_caching(cache_policy, get_after_set)
        return func
    return decorator

class QkitProperty:
    """
    A store of metadata for properties, which shall be handled in QKit.
    It does no range checking whatsoever.

    Do note, that this is a member of a class. You will need an object to perform calls.
    For this reason [WrappedProperty] exists.
    """

    QKIT_CACHE_NAME = "_qkit_cache"

    class ValueCache:
        last_value: Any
        dirty: bool

    # no longer implemented: name, channels, doc
    def __init__(self, *args, **kwargs):
        """
        See [QkitProperty.PropertyMetadata] for what the kwargs mean.
        """
        if len(args) != 0:
            raise SyntaxError("Missing Braces on '@QkitProperty()'!")
        self.metadata = PropertyMetadata(**kwargs)
        self.cache_policy = CachePolicy.NONE
        self.get_after_set = False

    def __call__(self, func):
        assert not not func.__doc__, f"Documentation is missing on {func.__name__}!"
        self.__doc__ = func.__doc__
        self.metadata.documentation = func.__doc__
        self.metadata.name = func.__name__
        self.fget = func
        return self

    def __get__(self, obj, objtype=None, query=True):
        """
        Emulate a property-getter. Default call to _get_logic.
        """
        return self._get_logic(obj)

    def _get_logic(self, obj, query=True, **kwargs):
        """
        Gets the value either from the wrapped getter or the cache, depending on [cache_policy]
        Used to access the cache if required. Used for internal calls.

        We ignore all further kwargs, as they may result in incompatibilities.
        """
        if obj is None:
            return 
        
        # We need to determine if we hit the cache or call the actual device.
        hit_cache = False # Per default we can not assume that caching is a good idea. Let's be pessimistic.
        if not query and not self.cache_policy == CachePolicy.ALWAYS_REFRESH: # Automated QKit-Get call. Hit cache if we are not force to always refresh.
            hit_cache = True
        elif query and self.cache_policy == CachePolicy.SOFT_GET: # We have no getter. Always go to cache
            hit_cache = True
        else:
            hit_cache = False

        if hit_cache:
            return self._qkit_cache(obj).last_value
        else: # Default query. Call getter.
            new_value = self.fget(obj)
            self._qkit_cache(obj).last_value = new_value
            return new_value

    def setter(self, fset):
        """
        Decorator to create the 
        """
        sig = inspect.signature(fset)
        assert len(sig.parameters) == 2, "Missing parameters in setter, expected (self, arg) or similar!"
        self.fset = fset
        return self

    def __set__(self, obj, value):
        """
        Wraps the setting of the value.
        """
        if self.fset is None:
            raise AttributeError("Can't set attribute!")
        if inspect.isfunction(value):
            raise ValueError("Likely bug detected: Attempting to set a QkitProperty to a function!")
        assert self.metadata.check_arg(value), "Argument validity check failed!"
        
        retval = self.fset(obj, value)
        # Update cache either with value or __get__
        if self.get_after_set:
            self._get_logic(obj, query=True)
        else:
            self._qkit_cache(obj).last_value = value

        return retval
        

    def set_cache_policy(self, cache_policy: CachePolicy, get_after_set: bool):
        self.cache_policy = cache_policy
        self.get_after_set = get_after_set


    def _qkit_cache(self, obj) -> ValueCache:
        """
        A consistent way to get the [obj]'s qkit value cache and a member of it.
        Creates the cache, should it not already exist.
        Creates a ValueCache instance for this property, should none exist.
        """
        if not hasattr(obj, QkitProperty.QKIT_CACHE_NAME):
            setattr(obj, QkitProperty.QKIT_CACHE_NAME, {})
        cache: dict = getattr(obj, QkitProperty.QKIT_CACHE_NAME)
        name = self.metadata.name
        if not name in cache:
            cache[name] = QkitProperty.ValueCache()
        return cache[name]

    @classmethod
    def from_property(cls, prop: property, metadata: PropertyMetadata) -> QkitProperty:
        """
        Convert a regular python property into a QkitProperty.
        Used to install a shim to notify us if regular calls are made.
        """
        instance = cls()
        instance.fget = prop.fget
        instance.fset = prop.fset
        instance.metadata = metadata
        return instance

    @staticmethod
    def get_metadata(obj, property_name: str) -> PropertyMetadata:
        """
        If the key is a QkitProperty of obj return its Metadata.
        Otherwise return None. 
        """
        try:
            candidate = type(obj).__dict__[property_name]
        except KeyError:
            return None
        if isinstance(candidate, QkitProperty):
            return candidate.metadata
        return None

class ObjectTreeVisistor(ABC):
    """
    A visitor visiting class members to register them as properties and functions of 
    """
    def __init__(self, instrument: ModernInstrument) -> None:
        self.instrument = instrument

    @abstractmethod
    def visit_instance(self, obj, path=[]):
        """
        Visists an object and disects it into usable pieces.
        If some method is intended to be used as a function, call [visit_function]
        If some property is inteded to be used as a property, call [visit_property]
        If some array or iterable and indexable structure is channel-like, call [visit_channel]
        """
        pass
    
    def get_property_metadata(self, obj: Any, name: str, prop: property) -> PropertyMetadata:
        """
        Provides an instance of [PropertyMetadata] for a discovered property if we were
        not able to automatically derive it from [@QkitProperty].
        """
        raise NotImplementedError("Unimplemented 'get_property_metadata'!")

    def wrap_property(self, obj: Any, name: str, prop: property) -> ModernInstrument.WrappedProperty:
        raise TypeError(f"Unsupported property type '{type(prop)}! Override 'wrap_property' missing!'")

    def wrap_function(self, obj: Any, name: str, func: Callable) -> ModernInstrument.WrappedFunction:
        raise TypeError(f"Unsupported property type '{type(func)}'! Override 'wrap_function' missing!")

    @staticmethod
    def derive_name_from_path(path: List(str)) -> str:
        return "_".join(map(lambda it: str(it), path))

    def visit_property(self, obj, name, prop, path=[]):
        """
        Handles a property [prop] of [obj] with the name [name].
        Registers it in discovered properties.
        """
        property_name = self.derive_name_from_path(path + [name])
        if isinstance(prop, QkitProperty):
            wrapped = ModernInstrument.WrappedProperty.from_qkit_property(obj, prop)
        elif isinstance(prop, property):
            wrapped = ModernInstrument.WrappedProperty(
                getter=lambda: property.__get__(obj),
                setter=lambda value: property.__set__(obj, value),
                metadata=self.get_property_metadata(obj, name, prop)
            )
        else:
            wrapped = self.wrap_property(obj, name, prop)
        self.instrument._parameters[property_name] = wrapped
        group = wrapped.get_metadata().group
        if group in self.instrument._parameter_groups:
            self.instrument._parameter_groups[group].append(wrapped)
        else:
            self.instrument._parameter_groups[group] = [wrapped]

    def visit_function(self, obj, name, func, path=[]):
        """
        Handles a function [func] of [obj] with the name [name].
        Registers it in discovered functions.
        """
        property_name = self.derive_name_from_path(path + [name])
        if inspect.isfunction(func) and hasattr(func, "_qkit_function"):
            wrapped = ModernInstrument.WrappedFunction.from_qkit_function(obj, func)
        else:
            wrapped = self.wrap_function(obj, name, func)
        self.instrument._functions[property_name] = wrapped

class QkitNativeVisitor(ObjectTreeVisistor):
    """
    This visitor is good for home-made drivers. It assumes that all properties we want to register
    have been properly marked as QkitProperty. It will not support other properties.

    You can take inspiration from this approach to write your own visitor to shim regular properties
    with [QkitProperty.from_property()] and setting the __dict__ entry.
    """

    def visit_instance(self, obj, path=[]):
        # Check for properties
        setattr(obj, "_qkit_path", path)
        setattr(obj, "_qkit_instrument", self.instrument)
        cls_defined = type(obj).__dict__
        for candidate in cls_defined:
            if candidate.startswith("_"):
                continue
            prop = cls_defined[candidate]
            if isinstance(prop, QkitProperty):
                # This is a QkitProperty. We can wrap it like normal.
                self.visit_property(obj, candidate, cls_defined[candidate], path)
            elif inspect.isfunction(prop) and hasattr(prop, "_qkit_function"):
                # This is a QkitFunction. Wrap it like a function.
                self.visit_function(obj, candidate, cls_defined[candidate], path)
        obj_defined = obj.__dict__
        for candidate in obj_defined:
            if candidate.startswith("_"):
                continue
            prop = obj_defined[candidate]
            if hasattr(prop, "__len__") and hasattr(prop, "__getitem__"):
                # This is a list, which might be channel-like
                for i in range(len(prop)):
                    self.visit_instance(prop[i], path = path + [candidate, str(i)])

class PropertyMetadata:
    """
    Stores metadata about a property. Is read out by [ObjectTreeVisistor]
    to create wrapped properties for [ModernInstrument]
    """

    def __init__(self, type=None, arg_checker=None, tags=[], subscribe=[], units=None, group=None) -> None:
        """
        Optional Arguments:
        type: float, int, str, np.ndarray, ...
        arg_checker: Checks argument for validity (minimum, maximum, steps) and returns either True (valid) or False
        tags: Tags associated with the parameter
        subscribe: A set of parameter names whose change triggers an update of this one 
        units: A string describing this properties units
        group: A string name of a group this property belongs to.
        """
        self.type = type
        self.arg_checker = arg_checker
        self.tags = tags
        self.subsriptions = subscribe
        self.units = units
        self.name = None
        self.documentation = None
        self.group = group

    def check_arg(self, arg):
        if self.arg_checker == None:
            return True
        else:
            return self.arg_checker(arg)

    def __repr__(self) -> str:
        import textwrap
        return textwrap.dedent(f"""
        Property {self.name}({self.type} {self.units}): tagged {self.tags} in group {self.group}
        listens to {self.subsriptions}
        argument checked: {not self.arg_checker == None}
        Documentation: """) + textwrap.dedent(self.documentation)



# zmq based info_service client
# HR@KIT/2017

import qkit
import zmq
from qkit.core.lib.com.signals import SIGNALS
# reverse the SIGNAL
RSIGNALS = dict((v,k) for k,v in SIGNALS.iteritems())




class info_client(object):
    def __init__(self,host = None, port = None):
        
        if host:
            self.host = host
        else:
            self.host = qkit.cfg.get('info_host','localhost')
        if port:
            self.port = port
        else:
            self.port = qkit.cfg.get('info_port',5600)
            
        self.start_client(self.port,self.host)
        self.start_poll()
        
    def start_client(self,port,host):
        print port, host
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.SUB)

        self.socket.connect ("tcp://%s:%s" %(host, port))
        
    def start_poll(self):
        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN)
        
    def listen_to(self,topic):
        "hook up to topic (defined in signals.py)"
        if SIGNALS.has_key(topic):
                sig = SIGNALS.get(topic)
                self.socket.setsockopt(zmq.SUBSCRIBE, str(sig).encode())
    def listen_to_all(self):
        for topic in SIGNALS.keys():
            self.listen_to(topic)
            
    def get_message(self,readable_topic = False):
        "wait until a message is available and return it"
        tid,message =self.socket.recv().split(":",1)
        if readable_topic:
            return RSIGNALS.get(int(tid)), message
        else:
            return tid, message
        
    def print_message(self):
        "wait until a message is available and print it"
        print(self.get_message(readable_topic = True))
        

if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        host =  int(sys.argv[1])
        port =  int(sys.argv[2])
        ifc = info_client(host,port)
    else:
        ifc = info_client()
    ifc.listen_to_all()
    for i in range(10):
        ifc.print_message()



# load_awg.py
# started by M. Jerger and adapted by AS, JB


# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit
import numpy as np
import logging
from qkit.gui.notebook.Progress_Bar import Progress_Bar
import gc


def update_sequence(ts, wfm_func, sample, iq = None, loop = False, drive = 'c:', path = '\\waveforms', reset = True, marker=None, markerfunc=None, ch2_amp = 2,chpair=1,awg= None, show_progress_bar = True):
    '''
        set awg to sequence mode and push a number of waveforms into the sequencer
        
        inputs:
        
        ts: array of times, len(ts) = #sequenzes
        wfm_func: waveform function usually generated via generate_waveform using ts[i]; this can be a tuple of arrays (for channels 0,1, heterodyne mode) or a single array (homodyne mode)
        sample: sample object
        
        iq: Reference to iq mixer instrument. If None (default), the wfm will not be changed. Otherwise, the wfm will be converted via iq.convert()
        
        marker: marker array in the form [[ch1m1,ch1m2],[ch2m1,ch2m2]] and all entries arrays of sample length
        markerfunc: analog to wfm_func, set marker to None when used
        
        for the 6GS/s AWG, the waveform length must be divisible by 64
        for the 1.2GS/s AWG, it must be divisible by 4
        
        chpair: if you use the 4ch Tabor AWG as a single 2ch instrument, you can chose to take the second channel pair here (this can be either 1 or 2).
    '''
    qkit.flow.start()
    if awg==None:
        awg = sample.awg
    clock = sample.clock
    wfm_func2 = wfm_func
    if iq != None:
        wfm_func2 = lambda t, sample: iq.convert(wfm_func(t,sample))
    
    # create new sequence
    if reset:
        if "Tektronix" in awg.get_type():
            awg.set_runmode('SEQ')
            awg.set_seq_length(0)   #clear sequence, necessary?
            awg.set_seq_length(len(ts))
        elif "Tabor" in awg.get_type():
            awg.set('p%i_runmode'%chpair,'SEQ')
            awg.define_sequence(chpair*2-1,len(ts))
        
        #amplitude settings of analog output
        awg.set_ch1_offset(0)
        awg.set_ch2_offset(0)
        awg.set_ch1_amplitude(2)
        awg.set_ch2_amplitude(ch2_amp)

    #generate empty tuples
    wfm_samples_prev = [None,None]
    wfm_fn = [None,None]
    wfm_pn = [None,None]
    if show_progress_bar: p = Progress_Bar(len(ts)*(2 if "Tektronix" in awg.get_type() else 1),'Load AWG')   #init progress bar
    
    #update all channels and times
    for ti, t in enumerate(ts):   #run through all sequences
        qkit.flow.sleep()
        wfm_samples = wfm_func2(t,sample)   #generate waveform
        if not isinstance(wfm_samples[0],(list, tuple, np.ndarray)):   #homodyne
            wfm_samples = [wfm_samples,np.zeros_like(wfm_samples, dtype=np.int8)]
        
        for chan in [0,1]:
            if markerfunc != None:   #use markerfunc
                try:
                    if markerfunc[chan][0] == None:
                        marker1 = np.zeros_like(wfm_samples, dtype=np.int8)[0]
                    else:
                        marker1 = markerfunc[chan][0](t,sample)
                    
                    if markerfunc[chan][1] == None:
                        marker2 = np.zeros_like(wfm_samples, dtype=np.int8)[0]
                    else:
                        marker2 = markerfunc[chan][1](t,sample)
                
                except TypeError:   #only one markerfunc given
                    marker1, marker2 = np.zeros_like(wfm_samples, dtype=np.int8)
                    if chan == 0:
                        marker1 = markerfunc(t,sample)
                    
            elif marker == None:   #fill up with zeros
                marker1, marker2 = np.zeros_like(wfm_samples, dtype=np.int8)
            else: #or set your own markers
                c_marker1, c_marker2 = marker[chan]
                marker1 = c_marker1[ti]
                marker2 = c_marker2[ti]
            
            if "Tektronix" in awg.get_type():
                wfm_fn[chan] = 'ch%d_t%05d'%(chan+1, ti) # filename is kept until changed
                if len(wfm_samples) == 1 and chan == 1:
                    wfm_pn[chan] = '%s%s\\%s'%(drive, path, np.zeros_like(wfm_fn[0]))   #create empty array
                else:
                    wfm_pn[chan] = '%s%s\\%s'%(drive, path, wfm_fn[chan])
                awg.wfm_send(wfm_samples[chan], marker1, marker2, wfm_pn[chan], clock)
                
                awg.wfm_import(wfm_fn[chan], wfm_pn[chan], 'WFM')
                
                # assign waveform to channel/time slot
                awg.wfm_assign(chan+1, ti+1, wfm_fn[chan])
                
                if loop:
                    awg.set_seq_loop(ti+1, np.infty)
            elif "Tabor" in awg.get_type():
                if chan == 1:   #write out both together
                    awg.wfm_send2(wfm_samples[0],wfm_samples[1],marker1,marker2,chpair*2-1,ti+1)
                else: continue
            else:
                raise ValueError("AWG type not known")
            if show_progress_bar: p.iterate()

        gc.collect()

    if reset and "Tektronix" in awg.get_type():
        # enable channels
        awg.set_ch1_status(True)
        awg.set_ch2_status(True)
        awg.set_seq_goto(len(ts), 1)
        awg.run()
        awg.wait(10,False)
    elif reset and "Tabor" in awg.get_type():
        # enable channels
        #awg.preset()
        awg.set_ch1_status(True)
        awg.set_ch2_status(True)
    qkit.flow.end()
    if sample.__dict__.has_key('mspec'):
        sample.mspec.spec_stop()
        sample.mspec.set_segments(len(ts))
    return np.all([awg.get('ch%i_status'%i) for i in [1,2]])


# Yokogawa_GS200.py driver for Yokogawa GS200 DC voltag/current source
# adapted from GS820 driver, Lukas Gruenhaupt @KIT 07/2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import logging
import numpy
import struct
import time
from distutils.version import LooseVersion

class Yokogawa_GS200(Instrument):
    '''
    This is the driver for the Yokogawa GS200 multi channel source measure unit

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Yokogawa_GS200', address='<GBIP address>',
        reset=<bool>)
    '''

    def __init__(self, name, address, reset=False):
        '''
        Initializes the Yokogawa GS200, and communicates with the wrapper.

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
            reset (bool)     : resets to default values

        Output:
            None
        '''
        # Initialize wrapper functions
        logging.info(__name__ + ' : Initializing instrument Yokogawa_GS200')
        Instrument.__init__(self, name, tags=['physical'])

        # Add some global constants
        self._address = address
        self._visainstrument = visa.instrument(self._address)
        
        # Set termination characters (nessecarry for Ethernet communication)
        if  LooseVersion(visa.__version__) < LooseVersion("1.5.0"):            # pyvisa 1.4
            self._visainstrument.term_chars = ''
        else:                                                                  # pyvisa 1.8
            self._visainstrument.read_termination  = ''
            self._visainstrument.write_termination = ''

        # Add parameters to wrapper

#        self.add_parameter('source_function',
#            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
#            type=str, units='')

#        
#        self.add_parameter('operation_mode',
#            flags=Instrument.FLAG_SET,
#            type=str, units='')

        self.add_parameter('source_mode',
            flags=Instrument.FLAG_GETSET,
            type=str, units='')

        self.add_parameter('source_range',
            flags=Instrument.FLAG_GETSET ,
            units='', type=str)
           
        self.add_parameter('output',
            flags=Instrument.FLAG_GETSET ,
            units='', type=str)
        
        self.add_parameter('level', 
            flags=Instrument.FLAG_GETSET,
            type=float, units='')

#        self.add_parameter('voltage_protection', 
#            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
#            type=float, units='V')
#
#        self.add_parameter('current_protection', 
#            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
#            type=float, units='A')
#            
#        self.add_parameter('4W',
#            flags=Instrument.FLAG_GETSET ,
#            units='', type=str)
#
#        self.add_parameter('ramp_wait_time', 
#            flags=Instrument.FLAG_GET,
#            type=float, units = 's')

        # Add functions to wrapper
        
        self.add_function('reset')
        self.add_function('get_all')
        #self.add_function('set_range_auto')
        self.add_function('set_defaults')
        self.add_function('ramp_current')
        #self.add_function('ramp_ch2_current')

        
        if reset:
            self.reset()
        else:
            self.get_all()
            self.set_defaults()

# functions
            
    def reset(self):     
        '''
        Resets instrument to default values

        Input:
            None
    
        Output:
            None
        '''
        logging.debug(__name__ + ' : Resetting instrument')
        self._write('*RST')
        self.get_all()
      
    def get_all(self):
        '''
        Reads all relevant parameters from instrument

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Get all relevant data from device')
        
        #self._ask(':outp?')
        
        
        #self.get_sync()
        #self.get('source_function')
        
        #self.get_sync()
        


#        self.get('sense_range')
#        self.get('sense_mode')
#        self.get('source_trig')
#        self.get('sense_trig')
#        self.get('source_delay')
#        self.get('sense_delay')
#        self.get('4W')
        #self.get('level')
#        self.get('output')
           
    def do_get_output(self):
        '''
        Gets output state
        '''
        logging.debug('Get status of output')
        ans = self._ask(':outp?')
        
        if ans=='zero':    # TODO what is this
            return 'off'
        
        if int(ans):
            print('Attention: Output is set on; Be careful!')      #warning if output is on! No fast current changes if coil is attached
            return 'on'
        else:
            return 'off'
        
    def do_set_output(self, val):
        '''
        Sets outputs of channels in "on" or "off" position.
        
        '''
        if val in ['on','off']:
            logging.debug('Set output to %s' % val)
            self._write(':outp:stat %s' % val)
        else:
            logging.error('Invalid value %s' % val)
    
    def do_get_source_mode(self):
        
        logging.debug('Get source function mode')
        return self._ask(':sour:func?')
        
    def do_set_source_mode(self, mode):
        
        logging.debug('Set source function mode to CURR or VOLT')
        
        if mode in ['curr','volt']:
            
            logging.debug('Set mode to %s mode' % mode)
            self._write(':sour:func %s' % mode)
            
        else:
            logging.error('Invalid mode %s' % mode)
    
            
    
#    
#    def set_range_auto(self, channel):
#        '''
#        Switch autorange on.
#        '''
#        logging.debug('Set range to auto mode')
#        self._visainstrument.write(':chan%d:sens:rang:auto on' % channel)
#        
    def set_defaults(self):
        
        '''
        Set to driver defaults:
            
            TODO

        '''
        
        self._write(':syst:beep 0')
        self.set_source_mode('curr')
        self.set('source_range', '10e-3')
        
        
#        self.set('4W', 'off')

#        self.set('sense_mode', 'volt')
#        self.set('sense_range', '200mV')
#        self.set('source_delay', 15e-6)
#        self.set('sense_delay', 0.3)
#        self.set('source_trig', 'ext')
#        self.set('sense_trig', 'sour')
    

#        
# # parameters
#        
# 
#    def do_set_sync(self, val):
#        '''
#        Turns "on"/"off" the interchannel synchronization.
#        The first channel is always " The master", second - "the slave"
#        '''    
#        logging.debug('Sets the channels in synchronized regime')
#        if val in ['on','off']:
#            logging.debug('Set mode to %s' % val)
#            self._visainstrument.write(':sync:chan %s' % val)
#        else:
#            logging.error('Invalid value %s' % val)
#            
#    def do_get_sync(self):
#        '''
#        Gets synchronization mode.
#        '''
#        logging.debug('Gets synchronization mode')
#        ans = self._visainstrument.ask(':sync:chan?')
#        
#        if int(ans):
#            return 'on'
#        else:
#            return 'off'
#
    def do_set_source_range(self, val):
        '''
        sets source range to definite channel 
         
        Input:
            val (string)  :
            channel (int) : 
        
        Output:
        
        '''
        logging.debug('Set source range to %s' % val)
        self._set_func_par_value('sour', 'rang', val)
       
       

    def do_get_source_range(self):
        '''
        Get source range 
        '''
        logging.debug('Get source range')
        return self._get_func_par('sour', 'rang')
#        
#        
#    def do_set_sense_range(self, val, channel):
#        '''
#        Set sense range to the specified channel
#
#        Input:
#            val (string)      : 
#            channel (int)     : 
#
#        Output:
#            None
#        '''
#        logging.debug('Set sense range to %s' % val)
#        if (val == 'auto'):
#            self.set_range_auto(channel)
#        else:
#            self._set_func_par_value(channel, 'sens', 'rang', val)
#
#    def do_get_sense_range(self, channel):
#        '''
#        Get sense range for the current mode.
#        
#        Input:
#            channel(int): 
#        Output:
#            range (string) : 
#        '''
#        logging.debug('Get sense range')
#        return self._get_func_par(channel, 'sens', 'rang')
#        
#
#        
#    def do_set_sweep_mode(self, mode, channel):
#        '''
#        Set source mode to current regime.
#        '''    
#        if mode in ['fix']:
#            logging.debug('Set mode to %s sweep mode', mode)
#            self._set_func_par_value(channel, 'sour', 'mode', mode)
#        else:
#            logging.error('invalid sweep mode %s' % mode)
#        self.get_all()
#
#    def do_get_sweep_mode(self, channel):
#        '''
#        Get source mode 
#        
#        Input:
#            channel (int) : 
#        '''
#        logging.debug('Get source sweep mode')
#        return self._get_func_par(channel, 'sour', 'mode')
#
#    def do_set_sense_mode(self, mode, channel):
#        '''
#        Set sense mode to current regime.
#        '''    
#        if mode in ['curr','volt']:
#            logging.debug('Set sense mode to %s mode', mode)
#            self._set_func_par_value(channel, 'sens', 'func', mode)
#        else:
#            logging.error('invalid mode %s' % mode)
#        self.get_all()
#        
#    def do_get_sense_mode(self, channel):
#        '''
#        Get sense mode for the current mode.
#        
#        Input:
#            channel (int) : 
#        '''
#        logging.debug('Get sense mode')
#        return self._get_func_par(channel, 'sens', 'func')
#        
#    def do_set_source_trig(self, trig, channel):
#        '''
#        Set source trigger.
#        '''    
#        if trig in ['ext','sens', 'aux']:
#            logging.debug('Set source trig to %s trigger', trig)
#            self._set_func_par_value(channel, 'sour', 'trig', trig)
#        else:
#            logging.error('invalid trig %s' % trig)
#        self.get_all()
#
#    def do_get_source_trig(self, channel):
#        '''
#        Get source trigger
#        '''
#        logging.debug('Get source trigger')
#        return self._get_func_par(channel, 'sour', 'trig')
#
#
#    def do_set_sense_trig(self, trig, channel):
#        '''
#        Set sense trigger.
#        '''    
#        if trig in ['ext','sour', 'aux']:
#            logging.debug('Set sense trig to %s trigger', trig)
#            self._set_func_par_value(channel, 'sens', 'trig', trig)
#        else:
#            logging.error('invalid trig %s' % trig)
#        self.get_all()
#
#    def do_get_sense_trig(self, channel):
#        '''
#        Get sense trigger
#        
#        '''
#        logging.debug('Get sense trigger')
#        return self._get_func_par(channel, 'sens', 'trig')
#        
#    def do_set_source_delay(self, val, channel):
#        '''
#        Set source delay
#        '''
#        logging.debug('Set source delay to %s' % val)
#        self._set_func_par_value(channel, 'sour', 'del', val)
#
#    def do_get_source_delay(self, channel):
#        '''
#        Get source delay
#        '''
#        logging.debug('Get source delay')
#        return float(self._get_func_par(channel, 'sour','del'))
#        
#    def do_set_sense_delay(self, val, channel):
#        '''
#        Set sense delay
#        '''
#        logging.debug('Set sense delay to %s' % val)
#        self._set_func_par_value(channel, 'sens', 'del', val)
#
#
#    def do_get_sense_delay(self, channel):
#        '''
#        Get sense delay
#        '''
#        logging.debug('Get sense delay')
#        return float(self._get_func_par(channel, 'sens','del'))
#        
#    def do_set_4W(self, val, channel):
#        '''
#        Sets measurement mode to 4-wire or 2-wire mode.
#        Value should be on or off. 
#        "On" devotes to 4-wire mode. "Off" devotes to 2-wire mode.
#        In the instrument appropriate command is ":sens:rem on/off"
#        '''
#        if val in ['on','off']:
#            logging.debug('Set 4W to %s' % val)
#            self._set_func_par_value(channel, 'sens', 'rem', val)
#        else:
#            logging.error('Invalid value %s' % val)
#
#    def do_get_4W(self, channel):
#        '''
#        Gets measurement mode
#        '''
#        logging.debug('Get 4-wire measurement mode')
#        ans = self._get_func_par(channel, 'sens', 'rem')
#        
#        if int(ans):
#            return 'on'
#        else:
#            return 'off'
#        
    def do_set_level(self, val):
        '''
        Set measuring level
        '''
        logging.debug('Set measuring level')
        #if self.get('ch%d_output' % channel, query = False) == 'off':
        #    self.set('ch%d_output' % channel, 'on')
        
        self._write(':sour:lev %s' % val)
        
    def do_get_level(self):
        '''
        Get measuring level
        '''
        logging.debug('Get measuring level')
        return self._ask(':sour:lev?')
        
#
#    def do_get_value(self, channel):
#        '''
#        Gets measured value
#        '''
#        logging.debug('Get measured value')
#        return float(self._visainstrument.ask(':chan%d:fetc?' % channel))
#
#
    def ramp_current(self,target, step, wait=0.1, showvalue=False):
        
        '''
        Ramps the current starting from the actual value to a target value
        Attention: all values are given in A
        'step' determines the step size
        'wait' determines the sleep time after every step
        'showvalue' print current value - default is False
        '''
        
        start = self.get_level()
        if showvalue==True: print("{:g}mA".format(round(start*1e3, 3)))
        
        if(target < start): step = -step
        a = numpy.concatenate( (numpy.arange(start, target, step)[1:], [target]) )
        for i in a:
            if showvalue==True: print("{:g}mA".format(round(i*1e3, 3)), end=" ")
            self.set_level(i)
            time.sleep(wait)
            
# core communication
  
    def _write(self, msg):
        
        '''
        Sends a visa command <msg>
        
        Input:
            msg (str)
        Output:
            None
        '''
        return self._visainstrument.write(msg)
    
    
    def _ask(self, msg):
        '''
        Sends a visa command <msg> and returns the read answer <ans>
        
        Input:
            msg (str)
        Output:
            ans (str)
        '''
        return self._visainstrument.query(msg).rstrip()
        
        
        
    def _set_func_par_value(self, func, par, val):
        '''
        For internal use only!!
        Changes the value of the parameter for the function specified
        
        Input:
            ch (int)
            func (string) :
            par (string)  :
            val (depends) :
        
        Output:
            None
        '''
        string = ':%s:%s %s' %(func, par, val)
        logging.debug(__name__ + ' : Set instrument to %s' % string)
        self._visainstrument.write(string)
       
       

    def _get_func_par(self, func, par):
        '''
        For internal use only!!
        Reads the value of the parameter for the function specified
        from the instrument
        
        Input:
            ch (int)      :
            func (string) :
            par (string)  :
        
        Output:
            val (string) :
        '''
        string = ':%s:%s?' %(func, par)
        ans = self._visainstrument.query(string)
        logging.debug(__name__ + ' : ask instrument for %s (result %s)' % \
            (string, ans))
        return ans.lower()



import os
import random
import time
import weakref

class File():

    _files = []
    _dir = ''

    def __init__(self, path=None, mode='w+', **kwargs):
        '''
        Create a temporary file, optionally with a name given by <path>.

        kwargs are options that will be set as properties of this object.
        '''

        if path is None or len(path) == 0:
            self.name = self.create_name()
        else:
            self.name = path

        self._mode = mode
        self._file = open(self.name, mode)

        for key, val in kwargs.iteritems():
            setattr(self, key, val)

        # Keep a weak reference
        File._files.append(weakref.ref(self))

    def __del__(self):
        self.remove()

    def create_name(self):
        tstr = time.strftime('%H%M%S')
        i1 = random.randint(0, 0xffffffff)
        fn = os.path.join(self._dir, '%s_%d.tmp' % (tstr, i1))
        return fn

    def close(self):
        if self._file is not None:
            self._file.close()
        self._file = None

    def reopen(self, mode=None):
        '''
        Reopen the temporary file.
        If mode is None (default) the mode passed when creating the object
        will be used.
        '''

        if mode is None:
            mode = self._mode
        self._file = open(self.name, mode)

    def remove(self):
        try:
            if self.name != '':
                os.remove(self.name)
                del File._files[name]
        except Exception:
            pass
        self.name = ''

    def read(self, *args):
        return self._file.write(*args)

    def write(self, *args):
        return self._file.write(*args)

    def flush(self):
        return self._file.flush()

    def get_file(self):
        '''Return the underlying file object.'''
        return self._file

    @staticmethod
    def remove_all():
        '''Remove all temporary files created through the File class.'''
        for i in File._files:
            if i() is not None:
                i().remove()

    @staticmethod
    def set_temp_dir(dir):
        File._dir = dir




#!/usr/bin/env python2
# DAC_LTC2666_communication.py by JB@KIT 01/2018

'''
Communication script between Raspberry Pi and DAC Board based on SPI commands. Requires the python package spidev 
installed on the Raspberry.
Raspberry Pi 3 model pin configuration: MOSI: 19, MISO: 21, SCLK: 23, Chip select CE0: 24 (in use with dac0), CE1: 26 (dac1). 
DAC model: LTC2666-16. 

Internal reference: REF floating, RD bit = 0.

The SPI communication can be performed with a 24bit word or alternatively with a 32bit word, 
both of which allowing for echo readback verifying data transfer. 
At present, the 24bit variant is implemented.
'''

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import time,sys
import logging

import spidev

def bytes_from_nibbles(nibbles):
    '''
    Combine a list of 6 nibbles (1 nibble = 4bit) into a 24bit load sequence, consisting out of a list of 3 bytes.
    '''
    if len(nibbles) != 6: raise ValueError
    return [(nibbles[0]<<4)+nibbles[1],(nibbles[2]<<4)+nibbles[3],(nibbles[4]<<4)+nibbles[5]]
    

class DAC_LTC2666(object):
    '''
    SPI communication class. Objects are 8-channel DAC boards.
    '''
    def __init__(self, cs = 0, word32 = False):
        '''
        Inputs:
        - cs: chip select connection
        - word32: switch to the 32bit version by passing 'True', default: 24bit
        '''
        self.word32 = word32
    
        self.SPIBUS = cs   #chip select
        self.spi = spidev.SpiDev()
        self.spi.open(0,self.SPIBUS)
        
        self.init()
        
    def verify_connection(self):
        ''' send no operation command and validate the echo readback '''
        if not self.word32: logging.warning("Currently, the SPI communication is set up with a 24bit word. Echo readback possibly not supported.")
        no_op = bytes_from_nibbles([15,0,0,0,0,0])
        self.write(no_op)
        if self.write(no_op) == no_op:
            return True
        else:
            return False
        
    def init(self):
        '''
        Initializes to internal reference. 
        Currently, all channels are used with +-5V output span, which does not have to be set when 
        using a correct MSP jumper setting.
        '''
        #set to internal reference mode
        self.write(bytes_from_nibbles([7,0,0,0,0,0]))
        
        #set all channels to +- 5V output span, not needed when MSP jumpers correctly set
        #for i in range(8):
        #    self.spi.xfer(bytes_from_nibbles([6,i,0,0,0,2])
        
        #self.verify_connection()
        
    def write(self, byte_arr):
        '''
        Generic SPI write command. Returns the SPI echo readback.
        '''
        if self.word32:
            return self.spi.xfer([0]+byte_arr)
        else:
            return self.spi.xfer2(byte_arr)   #write value in input register
        
    def update_output(self,channel):
        '''
        Update output according to input register. Used to acknowledge the previous command, which is returned.
        '''
        return self.write(bytes_from_nibbles([1,channel,0,0,0,0]))

    def set_val(self, channel, val):
        '''
        Set value in 0..2**16-1, 16bit resolution.
        Returns True for successful readback, returns False if an error occurred in the readback.
        '''
        byte_arr = bytes_from_nibbles([0,channel,(val>>12) & 15,(val>>8) & 15,(val>>4) & 15,val & 15])
        self.write(byte_arr)
        #echo readback delivers the previous input word and can be used to verify data transfer
        readback = self.update_output(channel)
        if readback == byte_arr:
            return True
        else:
            logging.error("{:s}: SPI communication not acknowledged.".format(__name__))
            return False

    def set_voltage(self, channel, voltage):
        '''
        Set voltage in range -5V..5V, using self.set_val(...).
        '''
        val = int((voltage+5.)/10*2**16)
        if val < 0 or val >= 2**16:
            logging.error("{:s}: Value to be set out of range.".format(__name__))
            return False
        else:
            return self.set_val(channel,val)

                

if __name__ == "__main__":   #if executed as main (and not imported)
    time.sleep(0.1) 
    
    dac = DAC_LTC2666(cs = 0)
    if dac.verify_connection():
        for ch in range(8):
            dac.set_voltage(ch,0)
        print 'ok'
    else: print 'Connection not valid.'
 


# -*- coding: utf-8 -*-

# MMW@KIT
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
References
----------
https://pypi.org/project/pcf8575/

Installation
------------
sudo apt-get update
sudo apt-get install libffi-dev
pip install pcf8575
"""

from qkit.config.services import cfg
import time
import logging
import os
import json

try:
    from qkit.services.qplexkit.pcf8575 import PCF8575
except ModuleNotFoundError:
    # logging.warning('qplexkit: Module PCF8575 for I2C-expander not found. Use dummy instead that just supports software debugging but no control of the qplexkit at all.')
    class PCF8575(object):
        """
        This is a dummy class that supports software debugging but no control of the qplexkit and is used for the qkit-driver.
        """
        def __init__(self, i2c_bus_no, address):
            print(i2c_bus_no, address)
            self.port = 16 * [True]



class qplexkit(object):
    """
    Qplexkit is a DC multiplexer for low level and low noise transport measurements of micro and nano circuits at low
    temperatures. It uses MDSM connectors with 25 pins and enables to measure 12 different 4-wire experiments with the
    same four lines in switching between them by means of low temperature applicable latching relays. The control unit
    at room temperature uses a Raspberry Pi Model B+ and provides 20 control lines connected to a current source to
    switch the relays. The control lines not required by the multiplexer can be used to switch cryogenic relays, e.g. to
    switch power dividers on or off for low-ohmic experiments to achieve better SNRs.
    A driver is embedded to Qkit - a quantum measurement suite in python.


    ##########################################################################
    #                            Cryostat circuit                            #
    ##########################################################################

                                                      bias                sense                                                     ---------------------------
                                                 │           │         │           │                                                CURRENT DIVIDER & AMPLIFIER
                                                 │           │       ╔═╧═╗       ╔═╧═╗
                                                 │    ╔═══╗  │    ┌──╢18 ╟──┐ ┌──╢18 ╟──┐
                                                 ├─■■─╢17 ╟──┤    │  ╚═══╝  / \  ╚═══╝  │                                           bias lines: 1/10 current divider
                                                 │ o──╢   ║  │    │  ╔═══╗ /___\        │                                           sense lines: amplifier
                                                 │    ╚═══╝  │    └──╢18 ╟──┘ └──╢18 ╟──┘
                                                 █    (13)   █       ╚═╤═╝  (14) ╚═╤═╝
                                                 │           │         │           │
                                                 │           │         │           │
                                                 └─────╥─────┘         └─────╥─────┘
                                                       ║                     ║                                                      =================== MDM-connectors (separate PCBs)
                                                       ║                     ║                                                      -------------------
                                                     ╔═╩═╗                 ╔═╩═╗                                                    PROTECTIVE RESISTOR
                                                (11) ║15 ╠════════╦════════╣16 ║ (12)
                                                     ╚═╦═╝        ║        ╚═╦═╝
                                                   10Ω █          ║          █ 10Ω                                                  protective resistor during switching
                                                       ║          ║          ║
                                                       ╨          ║          ╨
                                                      GND         ║         GND                                                     --------------
                                                                ╔═╩═╗                                                               DC MULTIPLEXER
                                  ╔═════════════════════════════╣ 0 ╠═════════════════════════════╗                                 relay layer A: <lay> = 0
                                  ║                             ╚═══╝                             ║
                                  ║                              (0)                              ║
                                ╔═╩═╗                                                           ╔═╩═╗
                  ╔═════════════╣ 1 ╠═════════════╗                               ╔═════════════╣ 2 ╠═════════════╗                 relay layer B: <lay> = 1
                  ║             ╚═══╝             ║                               ║             ╚═══╝             ║
                  ║              (1)              ║                               ║              (2)              ║
                ╔═╩═╗                           ╔═╩═╗                           ╔═╩═╗                           ╔═╩═╗
          ╔═════╣ 3 ╠═════╗               ╔═════╣ 4 ╠═════╗               ╔═════╣ 5 ╠═════╗               ╔═════╣ 6 ╠═════╗         relay layer C: <lay> = 2
          ║     ╚═══╝     ║               ║     ╚═══╝     ║               ║     ╚═══╝     ║               ║     ╚═══╝     ║
          ║      (3)      ║               ║      (4)      ║               ║      (5)      ║               ║      (6)      ║
        ╔═╩═╗           ╔═╩═╗           ╔═╩═╗           ╔═╩═╗           ╔═╩═╗           ╔═╩═╗           ╔═╩═╗           ╔═╩═╗
      ╔═╣ 7 ╠═╗       ╔═╣ 8 ╠═╗       ╔═╣ 9 ╠═╗       ╔═╣10 ╠═╗       ╔═╣11 ╠═╗       ╔═╣12 ╠═╗       ╔═╣13 ╠═╗       ╔═╣14 ╠═╗     relay layer D: <lay> = 3
      ║ ╚═══╝ ║       ║ ╚═══╝ ║       ║ ╚═══╝ ║       ║ ╚═══╝ ║       ║ ╚═══╝ ║       ║ ╚═══╝ ║       ║ ╚═══╝ ║       ║ ╚═══╝ ║     lrel: logical relay number
      ║  (7)  ║       ║  (8)  ║       ║  ---  ║       ║  ---  ║       ║  (9)  ║       ║ (10)  ║       ║  ---  ║       ║  ---  ║     <self._lrel2prel[<lrel>]>: (physical relay number)
    ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐   ┌─╨─┐
    │ 0 │   │ 1 │   │ 2 │   │ 3 │   │ 4 │   │ 5 │   │ 6 │   │ 7 │   │ 8 │   │ 9 │   │10 │   │11 │   │12 │   │13 │   │14 │   │15 │   <exp>                   : logical experiment number
    └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘   └───┘
     (0)     (1)     (2)     (3)     (4)     ---     (5)     ---     (6)     (7)     (8)     (9)    (10)     ---    (11)     ---    <self._exps(<exp>)>      : (physical experiment number)
    0000    0001    0010    0011    0100    0101    0110    0111    1000    1001    1010    1011    1100    1101    1110    1111    bin(<exp>)               : equals path through relays ABCD (0:left, 1:right)

    DC MULTIPLEX
    Each logical latching relay (double solid lined) switches all four measurement lines to select the wanted experiment
    (single solid line). The logical relay numbers 9, 10, 11, 12 are dummy relays that do not exist physically, but
    simplify the control logic by using bit operations of this obtained symmetric setup.

    CURRENT DIVIDER
    Above the DC multiplexer, latching relays (double solid lined) may connect
    a pair of two current lines with a 10Ω resistance (black box). Together
    with the down-streamed cables, filters etc (illustrated as resistances
    above) this forms switchable current dividers whose attenuation factor
    depends on the ratio of these resistances and attenuates both signal and
    noise. Since the added Nyquist noise depends on the square root of the
    temperature, this low temperature current divider provides better signal to
    noise ratios compared to those operating at room temperature.


    ##########################################################################
    #                        Room temperature circuit                        #
    ##########################################################################

    control lines for multiplexer relays

        VP2     VP3     VP4     VP9    VP10    VP11    VP12    VP15    VP16    VP17    VP22    VP23    VP24            physical control line (equals MDM pinout)
        (0)     (1)     (2)     (3)     (4)     (5)     (6)     (7)     (8)     (9)    (10)    (11)    (12)            (logical control line)
         │       │       │       │       │       │       │       │       │       │       │       │       │
         ≈       ≈       ≈       ≈       ≈       ≈       ≈       ≈       ≈       ≈       ≈       ≈       ≈             connection to cryostat circuit
         │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗
     ...─┴─╢ 0 ╟─┴─╢ 1 ╟─┴─╢ 2 ╟─┴─╢ 3 ╟─┴─╢ 4 ╟─┴─╢ 5 ╟─┴─╢ 6 ╟─┴─╢ 7 ╟─┴─╢ 8 ╟─┴─╢11 ╟─┴─╢12 ╟─┴─╢15 ╟─┴─╢16 ╟─...   logical relay number (see Cryostat circuit)
           ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝
            (0)     (1)     (2)     (3)     (4)     (5)     (6)     (7)     (8)     (9)    (10)    (11)    (12)


    control lines for supplementary relays

        VP1     VP7     VP8    VP14    VP20    VP21    VP25            physical control line (equals MDM pinout)
       (13)    (14)    (15)    (16)    (17)    (18)    (19)            (logical control line)
         │       │       │       │       │       │       │
         ≈       ≈       ≈       ≈       ≈       ≈       ≈             connection to cryostat circuit
         │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗ │ ╔═══╗
     ...─┴─╢17 ╟─┴─╢18 ╟─┴─╢19 ╟─┴─╢20 ╟─┴─╢21 ╟─┴─╢22 ╟─┴─╢23 ╟─...   logical relay number (see Cryostat circuit)
           ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝   ╚═══╝
           (13)    (14)    (15)    (16)    (17)    (18)    (19)

    To safe relay switch lines for low temperature relays (double solid lined), they are arranged ring-shaped where each
    line is connected to the relay coil '+' on the one side and to the relay '-' on the other side. At room temperature
    they are floating by default and can be connected either with HIGH or LOW by two non-latching relays (single solid
    lined) that are controlled by GPIOs of a Raspberry Pi 3 Model B.
    """

    def __init__(self):
        """
        Initiates variables to default configuration and sets raspberry up.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self.cfg = cfg['qplexkit']
        ''' prepare raspberry pi '''
        # try:
        #     ## set board mode to Broadcom
        #     GPIO.setmode(GPIO.BCM)
        #     ## setup GPIO
        #     for pin in self._gpio.values():
        #          GPIO.setup(pin, GPIO.OUT)
        # except:
        #     logging.error('qplexkit: Cannot set up Raspberry Pi')
        #     raise RuntimeError('qplexkit: Cannot set up Raspberry Pi')
        ''' prepare I2C IO expander (PCF8575) '''
        self.pcf1 = PCF8575(i2c_bus_no=1, address=0x20)
        self.pcf2 = PCF8575(i2c_bus_no=1, address=0x21)
        self.pcf1.port[15] = False  # current source off (no inverted logic, as current source switches GND like the PCF8575)
        self.pcfs = {1: self.pcf1, 2: self.pcf2}
        ''' class variables '''
        self._switch_time = 10e-3
        ''' logical - physical relay '''
        self._lrel2prel = {0: 0,  # layer 0
                           1: 1, 2: 2,  # layer 1
                           3: 3, 4: 4, 5: 5, 6: 6,  # layer 2
                           7: 7, 8: 8, 9: None, 10: None, 11: 9, 12: 10, 13: None, 14: None,  # layer3
                           15: 11, 16: 12,  # relays for protective resistors
                           ### separate PCB ###
                           17: 13, 18: 14, 19: 15, 20: 16, 21: 17, 22: 18, 23: 19,  # supplementary relays
                           }  # logical relay to physical relay: physical relay number [logical relay number]
        self._prel2lrel = {val: key for key, val in self._lrel2prel.items()}  # physical relay to logical relay: logical relay number [physical relay number]
        del self._prel2lrel[None]
        self._lrel_num = len(self._lrel2prel)
        self._prel_num = len(self._prel2lrel)
        ''' logical - physical experiment '''
        self._lexp2pexp = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: None, 6: 5, 7: None,  # connector A
                           8: 6, 9: 7, 10: 8, 11: 9, 12: 10, 13: None, 14: 11, 15: None,  # connector B
                           }  # logical experiment to physical experiment: physical experiment number [logical experiment number]
        self._pexp2lexp = {val: key for key, val in self._lexp2pexp.items()}  # physical experiment to logical experiment: logical experiment number [physical experiment number]
        del self._pexp2lexp[None]
        ''' logical relay control line - physical relay control line '''
        self._lline2pline = {0: 2, 1: 3, 2: 4, 3: 9, 4: 10, 5: 11, 6: 12, 7: 15, 8: 16, 9: 17, 10: 22,  # multiplexer
                             11: 23, 12: 24,  # relays for protective resistors
                             ### separate PCP ###
                             13: 1, 14: 7, 15: 8, 16: 14, 17: 20, 18: 21, 19: 25,  # supplementary relays
                             }
        ''' logical relay control line - physical relay control line '''
        self._pline2port = {1: (2, 3),  # K_OnOff_VP1 - 2.p03
                            2: (1, 2),  # K_OnOff_VP2 - 1.p02
                            3: (1, 3),  # K_OnOff_VP3 - 1.p03
                            4: (1, 4),  # K_OnOff_VP4 - 1.p04
                            7: (2, 0),  # K_OnOff_VP7 - 2.p00
                            8: (2, 1),  # K_OnOff_VP8 - 2.p01
                            9: (1, 5),  # K_OnOff_VP9 - 1.p05
                            10: (1, 6),  # K_OnOff_VP10 - 1.p06
                            11: (1, 7),  # K_OnOff_VP11 - 1.p07
                            12: (1, 8),  # K_OnOff_VP12 - 1.p08
                            14: (2, 2),  # K_OnOff_VP14 - 2.p02
                            15: (1, 9),  # K_OnOff_VP15 - 1.p09
                            16: (1, 10),  # K_OnOff_VP16 - 1.p10
                            17: (1, 11),  # K_OnOff_VP17 - 1.p11
                            20: (2, 4),  # K_OnOff_VP20 - 2.p04
                            21: (2, 5),  # K_OnOff_VP21 - 2.p05
                            22: (1, 12),  # K_OnOff_VP22 - 1.p12
                            23: (1, 13),  # K_OnOff_VP23 - 1.p13
                            24: (1, 14),  # K_OnOff_VP24 - 1.p14
                            25: (2, 6),  # K_OnOff_VP25 - 2.p06
                            '24>2': (1, 15),  # K_VP24>VP2 - 1.p17
                            '25>7': (2, 7),  # K_VP25>VP7 - 2.p07
                            'I_src_OnOff': (1, 0),  # I_SRC_OnOff - 1.p00
                            'I_src_out_toggle': (1, 1),  # I_SRC_OUT_TOGGLE - 1.p01
        }
        ''' Code condition register with current settings '''
        self._ccr_file = self.cfg['ccr_file']
        try:
            self._ccr = self.read_ccr()
        except (IOError, ValueError) as e:
            logging.error(f'qplexkit: Cannot read condition code register: {e}')
            logging.info('qplexkit: Reset qplexkit and create new json-file')
            self._ccr = 0
            self._create_ccr()
            self.reset()

    def set_switch_time(self, val):
        """
        Sets duration of the current pulse to <val>.

        Parameters
        ----------
        val: float
            Duration of the current pulse for switching the relays (in seconds).

        Returns
        -------
        None
        """
        self._switch_time = val
        return

    def get_switch_time(self):
        """
        Gets duration of the current pulse <val>.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Duration of the current pulse for switching the relays (in seconds).
        """
        return self._switch_time

    def set_experiment(self, exp, protect=False, **kwargs):
        """
        Enables experiment <exp> by setting all relevant relays if necessary.

        Parameters
        ----------
        exp: int
            physical experiment number ∈ [0,11].
        protect: bool
            Measurement lines are connected using protective resistors to GND before and disconnected after changing to
            experiment <exp>. Default is False.
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        None
        """
        _ccr = kwargs.get('ccr', self._ccr)
        try:
            _lexp = self._pexp2lexp[exp]  # logical experiment number
            logging.info(f'Set experiment to {_lexp}({exp})')
            if protect:
                self.set_relay(rel=15, status=1)
                self.set_relay(rel=16, status=1)
            for lay in range(4):
                lrel = 2 ** lay - 1 + (_lexp >> 4 - lay)  # logical relay number
                lstatus = bool((_lexp >> 3 - lay) % 2)  # logical relay value
                if self.get_relay(rel=lrel, ccr=_ccr) ^ lstatus:  # if not already set correctly
                    self.set_relay(rel=lrel, status=lstatus, **kwargs)
            if protect:
                self.set_relay(rel=15, status=0)
                self.set_relay(rel=16, status=0)
        except (ValueError, KeyError) as e:
            logging.error(f'qplexkit: Cannot set experiment {exp}: {e}')
        return

    def get_experiment(self, **kwargs):
        """
        Sets experiment to <exp> in setting relevant relays if necessary.

        Parameters
        ----------
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        exp: int
            physical experiment number ∈ [0,11].
        """
        _ccr = kwargs.get('ccr', self._ccr)
        exp = 4 * ['0']
        for lay in range(4):
            exp[lay] = str((_ccr >> int(self._lrel_num + 1 - 2 ** lay -
                                        sum([2 ** (lay - a) * int(exp[a - 1]) for a in range(1, 4)]))) % 2)
        return self._lexp2pexp[int(''.join(exp), 2)]

    def set_current_divider(self, status, **kwargs):
        """
        Sets current divider to <status> by setting relevant the relay if necessary.

        Parameters
        ----------
        status: bool
            current divider state
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        None
        """
        raise NotImplementedError("current divider hardware not yet implemented")
        if bool(status) ^ self.get_relay(rel=17, **kwargs):
            return self.set_relay(rel=17, status=status, **kwargs)
        else:
            return

    def get_current_divider(self, **kwargs):
        """
        Gets <status> of current divider.

        Parameters
        ----------
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        status: bool
            current divider state
        """
        raise NotImplementedError("current divider hardware not yet implemented")
        return self.get_relay(rel=17, **kwargs)

    def set_amplifier(self, status, **kwargs):
        """
        Sets amplifier to <status> by setting relevant the relay if necessary.

        Parameters
        ----------
        status: bool
            amplifier state
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        None
        """
        raise NotImplementedError("voltage amplifier hardware not yet implemented")
        if bool(status) ^ self.get_relay(rel=18, **kwargs):
            return self.set_relay(rel=18, status=status, **kwargs)
        else:
            return

    def get_amplifier(self, **kwargs):
        """
        Gets <status> of amplifier.

        Parameters
        ----------
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        status: bool
            amplifier state
        """
        raise NotImplementedError("voltage amplifier hardware not yet implemented")
        return self.get_relay(rel=18, **kwargs)

    def set_pcf_port(self, pcf, port, status):
        """
        Sets the port <port> of the I2C IO-Expander PCF8675 <pcf> to <status>. Note the inverted logic.

        Parameters
        ----------
        pcf: int
            PCF8575 number
        port: int
            Port number
        status: bool
            port status

        Returns
        -------
        None
        """
        self.pcfs[pcf].port[15 - port] = status  # inverse order of pcf8575.py is compensated by <15 - port>
        return


    def get_pcf_port(self, pcf, port):
        """
        Gets the status of port <port> of the I2C IO-Expander PCF8675 <pcf>. Note the inverted logic.

        Parameters
        ----------
        pcf: int
            PCF8575 number
        port: int
            Port number

        Returns
        -------
        status: bool
            port status
        """
        return self.pcfs[pcf].port[15 - port]  # inverse order of pcf8575.py is compensated by <15 - port>

    def set_current_source_status(self, status):
        """
        Sets current source to <status> to switch relays.

        Parameters
        ----------
        status: bool
            current source state

        Returns
        -------
        None
        """
        self.set_pcf_port(*self._pline2port['I_src_OnOff'], status)  # note not inversed polarity
        return

    def get_current_source_status(self):
        """
        Gets the status of the current source to switch relays.

        Parameters
        ----------
        None

        Returns
        -------
        status: bool
            current source state

        """
        return self.get_pcf_port(*self._pline2port['I_src_OnOff'])  # note not inversed polarity

    def set_relay(self, rel, status, **kwargs):
        """
        Sets relay number <rel> to <status> in sending a current pulse of duration <switch_time> on the control
        corresponding to <rel> with polarity corresponding to <status>

        Parameters
        ----------
        rel: int
            logical relay number <rel> ∈ [0:16]
        status: bool
            relay polarity
        **kwargs:
            switch_time: float
                duration of voltage pulse to switch relays
            ccr: int
                condition code register <ccr> of relay states
            ccr_file: str
                json-file to write <ccr>

        Returns
        -------
        None
        """
        _switch_time = kwargs.get('switch_time', self._switch_time)
        ## physical relay
        if rel not in self._prel2lrel.values():
            raise ValueError(f'Can not set relay {rel}, as it does not exist physically, but is rather a logical dummy relay.')
        _prel = self._lrel2prel[rel]  # physical relay number
        ## left and right logical relay control lines
        if _prel <= 12:  # multiplexer
            _lline_l = _prel
            _lline_r = (_prel + 1) % 13  # mod 13 closes cycle for last multiplexer relay
        elif 12 < _prel <= 19:  # supplementary relays
            _lline_l = _prel
            _lline_r = ((_prel + 1 - 13) % 7) + 13  # mod 7 closes cycle for last supplementary relay and +/- 13 shifts by the 13 multiplexer relays
        ## physical relay control lines
        _pline_l = self._lline2pline[_lline_l]  # left physical control line (connected to relay pin 8)
        _pline_r = self._lline2pline[_lline_r]  # right physical control line (connected to relay pin 1)
        print(f'qplexkit: Set relay {_prel}({rel}) to {status} with line {_lline_l} (VP{_pline_l}) and line {_lline_r} (VP{_pline_r})')
        logging.info(f'qplexkit: Set relay {_prel}({rel}) to {status} with line {_lline_l} (VP{_pline_l}) and line {_lline_r} (VP{_pline_r})')
        ## I2C ports
        self.set_pcf_port(*self._pline2port[_pline_l], False)  # enable physical control line to relay pin 8
        self.set_pcf_port(*self._pline2port[_pline_r], False)  # enable physical control line to relay pin 1
        self.set_pcf_port(*self._pline2port['I_src_out_toggle'], status ^ (_prel % 13) % 2)  # status xor parity of physical relay (mod 13 shift supplementary relays to zero in order to use parity analogue)
        if _prel == 12:
            ## invert physical line 24 to match with line 2
            self.set_pcf_port(*self._pline2port['24>2'], False)
        elif _prel == 19:
            ## invert physical line 25 to match with line 7
            self.set_pcf_port(*self._pline2port['25>7'], False)
        ''' send current pulse '''
        self.set_current_source_status(True)  # enable current source
        time.sleep(_switch_time)
        self.set_current_source_status(False)  # disable current source
        ''' disable all I2C ports '''
        self.pcf1.port = [True] * 16
        self.pcf2.port = [True] * 16
        #for (pcf, port) in self._pline2port.values():
        #    self.set_pcf_port(pcf, port, True)
        self.set_current_source_status(False)  # disable current source
        ''' save changes in ccr '''
        self._set_ccr(rel=rel, status=status, **kwargs)
        self._write_ccr(timestamp=True, **kwargs)
        return

    def get_relay(self, rel, **kwargs):
        """
        Gets value <status> of relay <rel> from a given condition code register <ccr>

        Parameters
        ----------
        rel: int
            logical relay number <rel> ∈ [0:16]
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        status: bool
            relay polarity
        """
        return self.get_ccr(rel, **kwargs)

    def get_relays(self, **kwargs):
        """
        Gets statuses <status> of all relays from a given condition code register <ccr>

        Parameters
        ----------
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        status: bool
            relay polarity
        """
        return list(map(int, list(bin(kwargs.get('ccr', self._ccr))[2:].zfill(self._lrel_num + 1))))

    def _set_ccr(self, rel, status, **kwargs):
        """
        Sets condition code register <ccr> from a given relay <rel> with status <status>

        Parameters
        ----------
        rel: int
            logical relay number <rel> ∈ [0:16]
        status: bool
            relay polarity
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        None
        """
        _ccr = list(map(str, self.get_relays(**kwargs)))
        _ccr[rel] = str(int(status))
        self._ccr = int(''.join(_ccr), 2)
        return

    def get_ccr(self, rel, **kwargs):
        """
        Gets status <status> of relay <rel> from a given condition code register <ccr>

        Parameters
        ----------
        rel: int
            logical relay number <rel> ∈ [0:16]
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        status: bool
            relay polarity
        """
        return bool(int(kwargs.get('ccr', self._ccr) >> int(self._lrel_num - rel)) % 2)

    def _write_ccr(self, timestamp=True, **kwargs):
        """
        Writes a given condition code register <ccr> (and a current timestamp if wanted <timestamp>) to a json-file
        <ccr_file>. The format is for
        * timestamp = True: [{"ccr": <ccr>, "time": "yyyy-MM-dd HH:mm:ss"}]
        * timestamp = False: [<ccr>]

        Parameters
        ----------
        timestamp: bool
            Default is True.
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states
            ccr_file: str
                json-file to write <ccr>

        Returns
        -------
        None
        """
        try:
            _ccr = kwargs.get('ccr', self._ccr)
            _ccr_file = kwargs.get('ccr_file', self._ccr_file)
            with open(_ccr_file, 'r+') as f:
                try:
                    data = json.load(f)
                except ValueError as e:
                    logging.error(f'qplexkit: Cannot load json-file to append ccr: {e}')
                    data = []
                f.seek(0)
                if timestamp:
                    data.append({'ccr': _ccr, 'time': time.strftime('%Y-%m-%d %H:%M:%S')})
                else:
                    data.append(_ccr)
                json.dump(data, f)
        except IOError as e:
            logging.error(f'qplexkit: Cannot find json-file {_ccr_file}: {e}')
        return

    def read_ccr(self, n=-1, timestamp=False, **kwargs):
        """
        Reads the n-th condition code register <ccr> from a json-file <ccr_file>.

        Parameters
        ----------
        n: int
            List index in <ccr_file>
        timestamp: bool
            Default is False.
        **kwargs:
            ccr_file: str
                json-file to write <ccr>

        Returns
        -------
        ccr: int
            condition code register <ccr> of relay states
        """
        _ccr_file = kwargs.get('ccr_file', self._ccr_file)
        with open(_ccr_file, 'r+') as f:
            try:
                _ccr = json.load(f)[n]
                if isinstance(_ccr, dict):  # format of entry with timestamp: [{"ccr": <ccr>, "time": "yyyy-MM-dd HH:mm:ss"}]
                    if timestamp:
                        return _ccr['ccr'], _ccr['time']
                    else:
                        return _ccr['ccr']
                elif isinstance(_ccr, int):  # format of entry without timestamp: [<ccr>]
                    if timestamp:
                        raise AttributeError('No timestamp of requested ccr-entry available.')
                    else:
                        return _ccr
                else:
                    logging.error(f'qplexkit: Cannot handle data from json-file {_ccr_file}')
                    raise ValueError('Cannot read condition code register')
            except (ValueError, KeyError) as e:
                logging.error(f'qplexkit: Cannot find condition code register: {e}')
                raise ValueError('Cannot read condition code register')

    def _create_ccr(self, **kwargs):
        """
        Creates a json-file <ccr_file> and writes a condition code register <ccr>.

        Parameters
        ----------
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states
            ccr_file: str
                json-file to write <ccr>

        Returns
        -------
        None
        """
        _ccr = kwargs.get('ccr', self._ccr)
        _ccr_file = kwargs.get('ccr_file', self._ccr_file)
        if not os.path.isfile(_ccr_file):
            with open(_ccr_file, 'w+') as f:
                json.dump([{'ccr': _ccr, 'time': time.strftime('%Y-%m-%d %H:%M:%S')}], f)
        return

    def print_ccr(self, **kwargs):
        """
        Prints the state of all physical relays from a code condition register <ccr>.

        Parameters
        ----------
        **kwargs:
            ccr: int
                condition code register <ccr> of relay states

        Returns
        -------
        None
        """
        print('relay:\tstate')
        for (_lrel, _prel), _state in zip(self._lrel2prel.items(), self.get_relays(**kwargs)):
            if _prel is not None:
                print(f'{_prel}:\t{_state}')

    def reset(self):
        """
        Reset every relay to False and condition code register to 0

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        logging.info('qplexkit: Reset every relay to False')
        for prel, lrel in self._prel2lrel.items():  # iterate over all physical relays
            self.set_relay(lrel, 0)  # set logical relay number
        self._ccr = 0
        self._write_ccr()
        return




# gate_func.py
# adapted from earlier versions by Jochen Braumueller <jochen.braumueller@kit.edu>, 05/2016

# The gate function resets an AWG to the first waveform of a sequence for state==False and sends a trigger
# by raising a gate for state==True.
# Separate functions are supplied for Tektronix AWG and Tabor AWG

"""
usage and inputs:

 from qkit.measure.timedomain import gate_func
 or: import gate_func
 mspec.set_gate_func(gate_func.Gate_Func(awg = tawg, ni_daq = ftdidaq).gate_fcn)
 #pass the name of the AWG as imported to qtLAB (e.g. fastawg or tawg)
 #with the spectrum card wrapper virtual_measure_spec.py:
 mspec.set_gate_func(gfunc.gate_fcn)
"""

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import logging
import time
import qkit

class Gate_Func(object):

    def __init__(self, awg, pulser = None, ni_daq = None, ni_daq_ch = 'PFI0:0', sample=None):
        """
        initialize gate function
        inputs:
         - awg: qubit awg instrument object
         - pulser: instrument object (to send out pulse directly from pulser instead of triggering the pulser
         - ni_daq: ni_daq instrument object (in tunnel electronics rack)
         - ni_daq_ch: used channel of ni_daq
         - readout_tabor: tabor awg (2ch of 4ch) for readout
        outputs:
         -
        """
        self.awg = awg
        self._sample = sample
        if pulser:
            self.pulser = qkit.instruments.get(pulser)
            self._gate_high = self.pulser.set_mode_continuous
            self._gate_low = self.pulser.set_mode_gated
        else:
            self.pulser = None
        if ni_daq:
            self.ni_daq = qkit.instruments.get(ni_daq)
            self.ni_daq_ch = ni_daq_ch
            self._gate_high = lambda: self.ni_daq.digital_out(self.ni_daq_ch, 1)
            self._gate_low = lambda: self.ni_daq.digital_out(self.ni_daq_ch, 0)
        else:
            self.ni_daq = None
       
        self.selftriggered = (self.pulser == None and self.ni_daq == None)
        
        if self.selftriggered:
            self._gate_low  =  lambda: self.awg.set({'trigger_time':2, 'p1_trigger_time':2})
            self._gate_high = lambda: self.awg.set({'trigger_time':sample.T_rep, 'p1_trigger_time':sample.T_rep})
        
        if self.selftriggered and not ('Tabor' in self.awg.get_type()):
            raise ValueError("selftriggered mode (without pulser and nidaq) is currently only available for Tabor AWGs.")
        self.ni_daq_ch = ni_daq_ch
        
    def gate_fcn(self, state):
        """
        gate function
        inputs:
         - state: boolean, chooses one of the actions reset or trigger
        outputs:
         -
        """
        if state:   #switch gate to high, let the pulses run
            time.sleep(0.1)
            self._gate_high()
                
        else:   #reset AWG and wait for it
            self._gate_low()
            
            time.sleep(0.025)
            
            if 'Tektronix' in self.awg.get_type():
                self._reset_tek()
            elif 'Tabor' in self.awg.get_type():
                self._reset_tabor()
            else:
                logging.error('AWG unknown')
                raise ImportError
            
    def _reset_tek(self):
        """
        reset Tektronix AWG to first sequence
        """
        tries = 0
        while self.awg.get_seq_position() != 1:
            if tries > 30:
                logging.error('Tektronix AWG does not respond. Aborting.')
                raise ValueError('AWG did not reset to first waveform after %i iterations. There is something seriously going wrong. Check if AWG can enter run mode!'%tries)
            tries += 1
            self.awg.stop()
            self.awg.wait()
            self.awg.run()
            #self.awg.wait()
            time.sleep(0.05*seq_it)

        if tries > 4:
            print('Tektronix AWG reached first waveform after %d iterations.'%tries)
                
    def _reset_tabor(self):
        """
        reset Tabor AWG to first sequence
        """
        # workaround to intialize to first waveform and make the AWG ready
        if "p1_runmode" in self.awg.get_parameter_names():
            self.awg.set_p1_runmode('USER')
            self.awg.set_p2_runmode('USER')
            self.awg.set_p1_runmode('SEQ')
            self.awg.set_p2_runmode('SEQ')
        else:
            self.awg.set_runmode('USER')
            self.awg.set_runmode('SEQ')
        self.awg.fix(verbose=False)



import numpy as np
from zhinst.toolkit.control.drivers import SHFSG as _SHFSG
from qkit.drivers.ZHInst_Abstract import ZHInst_Abstract


class ZHInst_SHFSG(ZHInst_Abstract):
    def __init__(self, name, serialnumber, host="129.13.93.38",  **kwargs):
        super().__init__(name, **kwargs)
        self._shfsg = _SHFSG(name, serialnumber, host=host)
        self._shfsg.setup()
        self._shfsg.connect_device()
        
        # Iterate node tree for readable entries and hook them into QKit
        self.blacklist = ["system_fwlog", "features_code"]
        self.mount_api("node_dump_shfsg.txt", self._shfsg.nodetree)

        # Register readout methods

    
        



# This instrument transmits and receives raw ethernet packets to and from 
# a network interface. It is used to communicate with GHZdac devices.
# Markus Jerger <jerger@kit.edu>, 2011
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
import socket
import struct
import threading
try:
    import Queue
except ImportError:
    import queue as Queue

import types
import logging

class EthernetNode(Instrument):
    '''
    This instrument sends and receives ethernet packets (layer 2), to talk to GHZdac devices.

    Usage:
    Initialize with
    <name> = instruments.create('name', 'EthernetNode', interface='<interface name>', ethertype=<ethertype>)
    <interface name> = eth1
    <ethertype> = 0x0003 (ETH_P_ALL, receive all packets) or any other EtherType value
    '''

    # capture all packets ethertype
    ETH_P_ALL = 0x0003

    def __init__(self, name, interface, ethertype):
        '''
        Initializes the Oxford Instruments Kelvinox IGH Dilution Refrigerator.

        Input:
            name (string)    : name of the instrument
            interface (string) : label of the network interface used

        Output:
            None
        '''
        logging.debug(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._interface = interface
        self._ethertype = int(ethertype)
        self._queues = {}
        self._monitor = None # will receive all packets if set to a Queue object
        self._receiver = self.receiver(name = 'node_rx', args = (self))
        self._receiver.daemon = True
        self._updateLock = threading.Lock()

        #Add parameters
        self.add_parameter('interface', type=str, flags=Instrument.FLAG_GET)
        self.add_parameter('ethertype', type=int, flags=Instrument.FLAG_GET)
        self.add_parameter('MACs', type=types.ListType, flags=Instrument.FLAG_GET)
        self.add_parameter('rcvbuf', type=int, flags=Instrument.FLAG_GETSET)
        self.add_parameter('monitoring', type.bool, flags.Instrument.FLAG_GETSET)

        # Add functions
        self.add_function('register')
        self.add_function('unregister')
        self.add_function('monitor')
        self.add_function('send')

        # open socket
        self._init()


    class listener(threading.Thread):
        '''
        listen for incoming connections
        '''

    class receiver(threading.Thread):
        '''
        receives network packets and sorts them into the appropriate queue
        '''
        def run(self, node):
            while(True):
                # receive packet
                (packet, sender) = node._socket.recvfrom(2048)
                # sort into queue
                node._updateLock.acquire(blocking = True)
                queue = node._queues.get(sender)
                if(queue != None)
                    queue.put(packet)
                if(node._monitor != None):
                    node._monitor.put((sender, packet))
                node._updateLock.release()

    def _init(self):
        # open receiving socket
	    if(self._ethertype == self.ETH_P_ALL):
            logging.info(__name__ + ' : EtherType set to ETH_P_ALL, this may generate high CPU load')
	    self._socket = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, self._ethertype)
        self._socket.bind((self._interface, self._ethertype))

        # determine default buffer size
        self.get_rcvbuf()

    def _MAC(self, MACin):
        '''
            convert mac address to byte string
            expected input format is 01:23:45:67:89:AB
        '''
        # split input and convert the parts to ints
        MAC = [int(byte, base = 16) for byte in str(MACin).split(':')]
        # merge the parts into a byte string
        return struct.pack('>6B', *MAC)

    def register(self, MAC):
        '''
        register a new DAC device with the server

        Input:
            MAC address
        Output:
            queue object
        '''
        MACkey = self._MAC(MAC)
        self._updateLock.acquire(blocking = True)
        if(not self._queues.has_key(MACkey)):
            logging.info(__name__ + ' : registering new device %s.'%MAC)
            self._queues[MACkey] = Queue.Queue()
        self._updateLock.release()
        return self._queues[MACkey]

    def unregister(self, MAC):
        '''
        unregister a DAC device from the server

        Input:
            MAC address
        Output:
            none
        '''
        MACkey = self._MAC(MAC)
        self._updateLock.acquire(blocking = True)
        if(self._queues.has_key(MACkey)):
            logging.info(__name__ + ' : unregistering device %s.'%MAC)
            self._queues.pop(MACkey)
        self._updateLock.release()

    def monitor(self):
        '''
        return monitor Queue
        '''
        return self._monitor

    def send(self, MAC, data)
        MACkey = self._MAC(MAC)
        self._socket.send(data)


    def do_get_interface(self):
        ''' return name of the network interface used '''
        return self._interface

    def do_get_ethertype(self):
        ''' return packet type received '''
        return self._ethertype

    def do_get_MACs(self):
        ''' return list of registered devices '''
        return self._queues.keys()

    def do_set_rcvbuf(self, bufsize):
        ''' set the receive buffer size of the socket; note that the size of the individual receive queues is unrelated '''
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, int(bufsize))

    def do_get_rcvbuf(self):
        ''' get the receive buffer size of the socket '''
        return self._socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)

    def do_set_monitoring(self, status):
        ''' enable or disable monitoring '''
        if(status == True):
            logging.debug(__name__ + ' : monitoring enabled')
            self._monitor = threading.Queue()
        else:
            logging.debug(__name__ + ' : monitoring disabled')
            self._monitor = None

    def do_get_monitoring(self):
        ''' check if monitoring is enabled '''
        return self._monitor != None



# Agilent_E8257D.py class, to perform the communication between the Wrapper and the device
# Pieter de Groot <pieterdegroot@gmail.com>, 2008
# Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#import types
import logging

from qkit import visa

from qkit.core.instrument_base import Instrument
from qkit.drivers.AbstractMicrowaveSource import AbstractMicrowaveSource


class Keysight_N5173B(AbstractMicrowaveSource):
    '''
    This is the driver for the Keysight N5173B Signal Genarator

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Keysight_N5173B', address='<GBIP address>, reset=<bool>')
    '''
    def __init__(self, name, address, reset=False):
        '''
        Initializes the Keysight_N5173B, and communicates with the wrapper.

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
            reset (bool)     : resets to default values, default=False
        '''
        logging.info(__name__ + ' : Initializing instrument Keysight_N5173B')
        super().__init__(name, tags=['physical'])

        self._address = address
        self._visainstrument = visa.instrument(self._address)
        if (reset):
            self.reset()
        else:
            self.get_all()


    def do_get_power(self):
        '''
        Reads the power of the signal from the instrument

        Input:
            None

        Output:
            ampl (float) : power in dBm
        '''
        logging.debug(__name__ + ' : get power')
        return float(self._visainstrument.query('POW:AMPL?'))

    def do_set_power(self, amp):
        '''
        Set the power of the signal

        Input:
            amp (float) : power in dBm

        Output:
            None
        '''
        logging.debug(__name__ + ' : set power to %f' % amp)
        self._visainstrument.write('POW:AMPL %s' % amp)
    
    #def do_get_phase(self):
    #    '''
    #    Reads the phase of the signal from the instrument
    #
    #    Input:
    #        None
    #
    #    Output:
    #        phase (float) : Phase in radians
    #    '''
    #    logging.debug(__name__ + ' : get phase')
    #    return float(self._visainstrument.query('PHASE?'))

    #def do_set_phase(self, phase):
    #    '''
    #    Set the phase of the signal
    #
    #    Input:
    #        phase (float) : Phase in radians
    #
    #    Output:
    #        None
    #    '''
    #    logging.debug(__name__ + ' : set phase to %f' % phase)
    #    self._visainstrument.write('PHASE %s' % phase)

    def do_get_frequency(self):
        '''
        Reads the frequency of the signal from the instrument

        Input:
            None

        Output:
            freq (float) : Frequency in Hz
        '''
        logging.debug(__name__ + ' : get frequency')
        return float(self._visainstrument.query('FREQ:CW?'))

    def do_set_frequency(self, freq):
        '''
        Set the frequency of the instrument

        Input:
            freq (float) : Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : set frequency to %f' % freq)
        self._visainstrument.write('FREQ:CW %s' % freq)

    def do_get_status(self):
        '''
        Reads the output status from the instrument

        Input:
            None

        Output:
            status (string) : 'On' or 'Off'
        '''
        logging.debug(__name__ + ' : get status')
        return bool(int(self._visainstrument.query('OUTP?')))


    def do_set_status(self, status):
        '''
        Set the output status of the instrument

        Input:
            status (string) : 'On' or 'Off'

        Output:
            None
        '''
        logging.debug(__name__ + ' : set status to %s' % status)
        
        if status == True:
            self._visainstrument.write('OUTP ON')
        elif status == False:
            self._visainstrument.write('OUTP OFF')
        else:
            raise ValueError('set_status(): can only set True or False')


    def reset(self):
        logging.info(__name__ + ' : resetting instrument')
        self._visainstrument.write('*RST')
        self.get_all()



from abc import ABC, abstractmethod
from importlib.resources import path
from typing import Union, Any, Callable
from matplotlib.pyplot import isinteractive
import inspect

import numpy as np
import logging

from zhinst.toolkit.control.node_tree import NodeList, Parameter
from zhinst.toolkit.interface import LoggerModule
from qkit.core.instrument_basev2 import ModernInstrument, QkitProperty, ObjectTreeVisistor, PropertyMetadata, CachePolicy
SETTING_TAG = 'setting'

class ZHInstQkitProperty(QkitProperty):

    def __init__(self, param: Parameter, name):
        self.cache_policy = CachePolicy.NONE
        self.get_after_set = False
        self._param = param

        def getter(instance, param=param):
            return param()

        self.fget = getter

        try:
            test = self.fget(param)
        except LoggerModule.ToolkitConnectionError as tce:
            raise TypeError("Get not supported. Skip.", tce)
                
        access = param._properties # A set of strings describing access: Read, Write, Setting
        if 'Setting' in access:
            tags = [SETTING_TAG] 
        else:
            tags = []

        if 'Write' in access:
            def setter(instance, value, param=param):
                return param(value)
            self.fset = setter

        property_type = param._type

        if property_type == 'Double' and type(test) == np.float64:
            property_type = float
        elif (property_type == 'Integer' or property_type == 'Integer (enumerated)' or property_type == 'Integer (64 bit)') and type(test) == np.int64:
            property_type = int
        elif property_type == 'String' and type(test) == str:
            property_type = str
        elif property_type == 'ZIVectorData':
            if type(test) == np.ndarray:
                property_type = np.ndarray
            elif type(test) == str:
                property_type = str
            else:
                raise TypeError(f"Unsupported ZIVetorData type {type(test)} of {param._path}")
        else:
            raise TypeError(f"Documented type '{property_type}' of '{param._path}' with returned type {type(test)} is not supported!")

        unit = getattr(param, "_unit", param)
        
        self.metadata = PropertyMetadata(type = property_type, tags=tags, units=unit)
        self.metadata.name = name
        self.metadata.documentation = param._description

    def __call__(self, *args, **kwargs):
        """
        Overwrites the calls to the zhinst-driver to pass through qkit in order to allow caching
        and logging.
        """
        if len(args) == 0:
            return self.__get__(self._param)
        else:
            return self.__set__(self._param, args[0])


class ZHInstNodeVisitor(ObjectTreeVisistor):
    def visit_instance(self, obj, path=[]):
        if hasattr(obj, "parameters"):
            for param_name in obj.parameters:
                try:
                    self.visit_property(obj, param_name, getattr(obj, param_name), path = path)
                except TypeError as e:
                    logging.warn(f"Caught exception processing {param_name} at {path}: {e}")
                    continue
        
        if hasattr(obj, "nodes"):
            for node_name in obj.nodes:
                if node_name == "_parent":
                    continue
                self.visit_instance(getattr(obj, node_name), path = path + [node_name])
        
        if isinstance(obj, NodeList):
            for i, child in enumerate(obj):
                self.visit_instance(child, path = path + [i])


    def wrap_property(self, obj: Any, name: str, prop: Parameter) -> ModernInstrument.WrappedProperty:
        prop = ZHInstQkitProperty(prop, name)
        obj.__dict__[name] = prop
        return ModernInstrument.WrappedProperty.from_qkit_property(obj, prop)

    def wrap_function(self, obj: Any, name: str, func: Callable) -> ModernInstrument.WrappedFunction:
        raise TypeError(f"Unsupported property type '{type(func)}'! Override 'wrap_function' missing!")

    @staticmethod
    def _derive_channel_name(path):
        return "_".join(map(lambda it: str(it), path))

class ZHInst_Abstract(ABC, ModernInstrument):

    def __init__(self, name, **kwargs) -> None:
        super().__init__(name, **kwargs)
        if not hasattr(self, "blacklist"):
            self.blacklist = []
        self.add_tag(SETTING_TAG)

    def mount_api(self, node_tree):
        visitor = ZHInstNodeVisitor(self)
        visitor.visit_instance (node_tree)



# -*- coding: utf-8 -*-
"""
@author: andre.schneider@kit.edu / 2019
@author: micha.wildermuth@kit.edu / 2019
@license: GPL

This file provides some methods to quickly plot different datasets in a
matplotlib (QT) window. The data file to show can be easily selected in
the database view of qkit.fid

make sure to have qkit.fid enabled and %matplotlib qt executed beforehand
"""
import h5py
import json
import matplotlib.pyplot as plt
import numpy as np
from IPython.core.display import display, HTML

import qkit


class QuickPlot(object):
    def __init__(self, maximize=True):
        """
        Quickplot routine to scan through your measurement data.
        The maximize switch can be used to increase the width of
        the notebook view, giving you more space for the database table.
        """
        self.fig, self.ax = plt.subplots(2, num="Quickplot", clear=True)
        self.args, self.kwargs = (), {}
        self.m_type, self.ds_type, self.num_plots = None, None, None
        self.remove_offset_x_avg = False
        self.remove_offset_y_avg = False
        self.unwrap_phase = False
        if maximize:
            display(HTML("<style>.container { width:100% !important; }</style>"))
    
    def plot_selected_df(self, change):
        uuid = qkit.fid._selected_df.index[0]
        self.d = h5py.File(qkit.fid.h5_db[uuid], "r")
        try:
            self.m_type = json.loads(self.d["entry/data0/measurement"][0])["measurement_type"]
        except:
            self.m_type = None
        if self.m_type == "transport":
            self.overlays = self.d["entry/views/IV"].attrs[u"overlays"]+1
        else:
            self.overlays = None
        try:
            self.ds_type = self.d["entry/data0/"+{"spectroscopy": "amplitude", "transport": "v_0"}[self.m_type]].attrs["ds_type"]
        except:
            self.ds_type = None
        if self.args is () and self.kwargs == {}:
            self.num_plots = {"spectroscopy": 2,
                              "transport": (1, self.overlays, self.overlays)[self.ds_type-1]}[self.m_type]
        else:
            self.num_plots = len(self.args)+sum([1 if type(val) is str else len(val) for val in self.kwargs.values()])
        try:
            plt.clf()
            self.fig, self.ax = plt.subplots(self.num_plots, num="Quickplot", clear=True)
            if not np.iterable(self.ax):
                self.ax = [self.ax]
            self.fig.canvas.draw()
            self.ax[0].set_title(uuid + " | " + qkit.fid.h5_info_db[uuid]["name"])
            args = {}
            if self.m_type == "spectroscopy":
                if "amplitude_midpoint" in self.d["entry/data0"].keys():
                    self.plot_2D("amplitude_midpoint", "phase_midpoint")
                else:
                    if self.args == ():
                        args["ds"] = ["amplitude", "phase"]
                    else:
                        args["ds"] = list(self.args)
                    [self.plot_1D, self.plot_2D, self.plot_3D][self.ds_type - 1](args)
            elif self.m_type == "transport":
                if self.args == () and self.kwargs == {}:
                    if self.ds_type == 1:
                        args["view"] = ["IV"]
                    else:
                        args["ds"] = ["v_" + str(i) for i in range(self.overlays)]
                else:
                    if len(self.args) != 0:
                        args["ds"] = list(self.args)
                    if "ds" in self.kwargs.keys():
                        if len(argss["ds"]) != 0:
                            args["ds"] = list(self.kwargs["ds"])
                        else:
                            args["ds"].append(self.kwargs["ds"])
                    if "view" in self.kwargs.keys():
                        args["view"] = self.kwargs["view"] if type(self.kwargs["view"])==list else [self.kwargs["view"]]
                [self.plot_1D, self.plot_2D, self.plot_3D][self.ds_type - 1](args)
            else:
                print("No matching entries found for file %s" % uuid)
            self.fig.tight_layout()
        except Exception as e:
            print(e)
        finally:
            self.d.close()
            [ax.set_title("") for ax in self.ax[1:]]
            self.ax[0].autoscale_view()
            self.fig.canvas.draw()
    
    @staticmethod
    def si_prefix(data):
        switch = int(np.floor(np.log10(np.max(np.abs(np.array(data)))) / 3)) * 3
        try:
            return {
                -9: [1e-9, "nano", "n"],
                -6: [1e-6, "micro", "u"],
                -3: [1e-3, "milli", "m"],
                0 : [1, "", ""],
                3 : [1e3, "kilo", "k"],
                6 : [1e6, "Mega", "M"],
                9 : [1e9, "Giga", "G"]
            }[switch]
        except KeyError:
            return [1, "", ""]
    
    def plot_3D(self, args):
        def _plot(ax, key, val):
            if key == "view":
                for i in range(self.overlays):
                    x_url, d_url = self.d["entry/views/" + val].attrs[u"xy_" + str(i)].split(":")
                    si_x = self.si_prefix(self.d[x_url])
                    si_d = self.si_prefix(self.d[d_url])
                    x_index = int(self.d[x_url].shape[0] / 2)
                    y_index = int(self.d[x_url].shape[1] / 2)
                    ax.plot(self.d[x_url][x_index,y_index,:] / si_x[0], self.d[d_url][x_index,y_index,:] / si_d[0])
                    ax.set_xlabel("%s (%s%s)" % (self.d[x_url].attrs["name"], si_x[2], self.d[x_url].attrs["unit"]))
                    ax.set_ylabel("%s (%s%s)" % (self.d[d_url].attrs["name"], si_d[2], self.d[d_url].attrs["unit"]))
            elif key == "ds":
                if "/" not in val:
                    val = "entry/data0/" + val
                si_x = self.si_prefix(self.d[self.d[val].attrs["x_ds_url"]])
                si_y = self.si_prefix(self.d[self.d[val].attrs["y_ds_url"]])
                z_index = int(self.d[val].shape[2] / 2)
                data = self.d[val][:, :, z_index].T
                if self.unwrap_phase and val.split("/")[-1] == "phase":
                    data.T[~np.isnan(data.T)] = np.unwrap(data.T[~np.isnan(data.T)])
                if self.remove_offset_x_avg:
                    data -= np.nanmean(data, axis=1, keepdims=True)
                if self.remove_offset_y_avg:
                    data -= np.nanmean(data, axis=0, keepdims=True)
                ax.pcolormesh(self.d[self.d[val].attrs["x_ds_url"]][:self.d[val].shape[0]] / si_x[0], self.d[self.d[val].attrs["y_ds_url"]][:self.d[val].shape[1]] / si_y[0], data)
                ax.set_xlabel("%s (%s%s)" % (self.d[self.d[val].attrs["x_ds_url"]].attrs["name"], si_x[2], self.d[self.d[val].attrs["x_ds_url"]].attrs["unit"]))
                ax.set_ylabel("%s (%s%s)" % (self.d[self.d[val].attrs["y_ds_url"]].attrs["name"], si_y[2], self.d[self.d[val].attrs["y_ds_url"]].attrs["unit"]))
        ax_iter = iter(self.ax)
        for key, vals in args.items():
            for val in vals:
                _plot(next(ax_iter), key, val)

    def plot_2D(self, args):
        def _plot(ax, key, val):
            if key == "view":
                for i in range(self.overlays):
                    x_url, d_url = self.d["entry/views/" + val].attrs[u"xy_" + str(i)].split(":")
                    si_x = self.si_prefix(self.d[x_url])
                    si_d = self.si_prefix(self.d[d_url])
                    y_index = int(self.d[x_url].shape[0] / 2)
                    ax.plot(self.d[x_url][y_index,:] / si_x[0], self.d[d_url][y_index,:] / si_d[0])
                    ax.set_xlabel("%s (%s%s)" % (self.d[x_url].attrs["name"], si_x[2], self.d[x_url].attrs["unit"]))
                    ax.set_ylabel("%s (%s%s)" % (self.d[d_url].attrs["name"], si_d[2], self.d[d_url].attrs["unit"]))
            elif key == "ds":
                if "/" not in val:
                    val = "entry/data0/" + val
                si_x = self.si_prefix(self.d[self.d[val].attrs["x_ds_url"]])
                si_y = self.si_prefix(self.d[self.d[val].attrs["y_ds_url"]])
                data = self.d[val][:].T
                if self.unwrap_phase and val.split("/")[-1] == "phase":
                    data[~np.isnan(data)] = np.unwrap(data[~np.isnan(data)], axis=0)
                if self.remove_offset_x_avg:
                    data -= np.nanmean(data, axis=1, keepdims=True)
                if self.remove_offset_y_avg:
                    data -= np.nanmean(data, axis=0, keepdims=True)
                ax.pcolormesh(self.d[self.d[val].attrs["x_ds_url"]][:self.d[val][:].shape[0]] / si_x[0],
                              self.d[self.d[val].attrs["y_ds_url"]][:self.d[val][:].shape[1]] / si_y[0],
                              data)
                ax.set_xlabel("%s (%s%s)" % (self.d[self.d[val].attrs["x_ds_url"]].attrs["name"], si_x[2], self.d[self.d[val].attrs["x_ds_url"]].attrs["unit"]))
                ax.set_ylabel("%s (%s%s)" % (self.d[self.d[val].attrs["y_ds_url"]].attrs["name"], si_y[2], self.d[self.d[val].attrs["y_ds_url"]].attrs["unit"]))
        ax_iter = iter(self.ax)
        for key, vals in args.items():
            for val in vals:
                _plot(next(ax_iter), key, val)
    
    def plot_1D(self, args):
        def _plot(ax, key, val):
            if key == "view":
                for i in range(self.overlays):
                    x_url, d_url = self.d["entry/views/"+val].attrs[u"xy_" + str(i)].split(":")
                    si_x = self.si_prefix(self.d[x_url])
                    si_d = self.si_prefix(self.d[d_url])
                    ax.plot(self.d[x_url][:] / si_x[0], self.d[d_url][:] / si_d[0])
                    ax.set_ylabel("%s (%s%s)" % (self.d[d_url].attrs["name"], si_d[2], self.d[d_url].attrs["unit"]))
                    ax.set_xlabel("%s (%s%s)" % (self.d[x_url].attrs["name"], si_x[2], self.d[x_url].attrs["unit"]))
            elif key == "ds":
                if "/" not in val:
                    val = "entry/data0/" + val
                si_x = self.si_prefix(self.d[self.d[val].attrs["x_ds_url"]])
                si_d = self.si_prefix(self.d[val])
                ax.plot(self.d[self.d[val].attrs["x_ds_url"]][:self.d[val].shape[0]] / si_x[0], self.d[val][:] / si_d[0])
                ax.set_ylabel("%s (%s%s)" % (self.d[val].attrs["name"], si_d[2], self.d[val].attrs["unit"]))
                ax.set_xlabel("%s (%s%s)" % (self.d[self.d[val].attrs["x_ds_url"]].attrs["name"], si_x[2], self.d[self.d[val].attrs["x_ds_url"]].attrs["unit"]))
        self.output = self.ax, args
        ax_iter = iter(self.ax)
        for key, vals in args.items():
            for val in vals:
                _plot(next(ax_iter), key, val)

    def show(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        qkit.fid.show()
        qkit.fid.grid.observe(self.plot_selected_df, names=["_selected_rows"])
        return qkit.fid.grid
